import itertools
import math
import sys

import numpy as np
import numpy.linalg
import functools

from pychemia import Structure, pcm_log
from pychemia.utils.mathematics import integral_gaussian
from pychemia.utils.periodic import atomic_number, covalent_radius, valence, atomic_symbol
from collections import OrderedDict


class StructureAnalysis:
    """
    The set of analysis provided by this class uses only structural information of one
    single structure. The kind of analysis includes coordination numbers, bonds, distances,
    hardness, fingerprints.
    Most of those analysis rely on a robust computation of inter-atomic distances.
    This class uses lazy evaluation for the distances and bonds to lower cpu and memory footprint.
    """

    def __init__(self, structure, supercell=(1, 1, 1), radius=50):
        """
        Takes one pychemia Structure object and will create a StructureAnalysis object. This object is computes the
        distances between all atoms in the unit cell with replicas extending up to a distance given by 'radius'.
        A supercell could be needed in cases where full connectivity of the entire crystal is needed.

        :param structure: A pychemia Structure object
        :param supercell: A supercell to be created from the Structure (Default=(1,1,1))
        :param radius: Maximal distance computed between two atoms
        """

        assert (isinstance(structure, Structure))
        if supercell != (1, 1, 1):
            self.structure = structure.supercell(supercell)
        else:
            self.structure = structure.copy()

        self._distances = None
        self._all_distances = None
        self._pairs = None
        self._supercell = supercell
        self._radius = radius
        # log.debug('Supercell : ' + str(self._supercell))
        # log.debug('Radius    : %7.2f' % self._radius)

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, value):
        assert (value > 0.0)
        if value != self._radius:
            self._distances = None
            self._pairs = None
            self._all_distances = None
        self._radius = value

    @property
    def distances(self):
        return self._distances

    def close_distances(self):
        """
        Computes the closest distances for all the atoms

        :return: (tuple) Return a bond's dictionary and distance's list
        """
        if self._pairs is None or self._distances is None:

            if self.structure.is_periodic:
                pcm_log.debug('Computing distances from scratch...')
                pairs_dict = {}
                distances_list = []
                index = 0
                for i, j in itertools.combinations(range(self.structure.natom), 2):
                    if index % 100 == 0:
                        pcm_log.debug('Computing distance between atoms %d and %d' % (i, j))
                    ret = self.structure.lattice.distance2(self.structure.reduced[i], self.structure.reduced[j],
                                                           radius=self.radius)
                    for k in ret:
                        if str(i) not in pairs_dict:
                            pairs_dict[str(i)] = [index]
                        else:
                            pairs_dict[str(i)].append(index)
                        if str(j) not in pairs_dict:
                            pairs_dict[str(j)] = [index]
                        else:
                            pairs_dict[str(j)].append(index)
                        ret[k]['pair'] = (i, j)
                        distances_list.append(ret[k])
                        index += 1
                for i in range(self.structure.natom):
                    ret = self.structure.lattice.distance2(self.structure.reduced[i], self.structure.reduced[i])
                    for k in ret:
                        if str(i) not in pairs_dict:
                            pairs_dict[str(i)] = [index]
                        else:
                            pairs_dict[str(i)].append(index)
                        ret[k]['pair'] = (i, i)
                        distances_list.append(ret[k])
                        index += 1
                self._pairs = pairs_dict
                self._distances = distances_list
            else:
                dm = self.structure.distance_matrix()
                dm += np.eye(len(dm)) * max(dm.flatten())
                pairs_dict = {}
                distances_list = []
                for i in range(self.structure.natom):
                    index = dm[:, i].argmin()
                    pairs_dict[str(i)] = [index]
                    distances_list.append(dm[index, i])
                self._pairs = pairs_dict
                self._distances = distances_list

        return self._pairs, self._distances

    def all_distances(self):

        if self._all_distances is None:
            ret = {}
            if not self.structure.is_periodic:
                dist_matrix = self.structure.distance_matrix()
            for i, j in itertools.combinations_with_replacement(range(self.structure.natom), 2):
                pair = (i, j)
                if self.structure.is_periodic:
                    ret[pair] = self.structure.lattice.distances_in_sphere(self.structure.reduced[i],
                                                                           self.structure.reduced[j],
                                                                           radius=self.radius)
                else:
                    ret[pair] = dist_matrix[i, j]

            self._all_distances = ret

        return self._all_distances

    def all_distances_by_species(self):

        all_distances = self.all_distances()
        ret = OrderedDict()

        atom_numbers = atomic_number(self.structure.species)
        a = list(itertools.combinations_with_replacement(atom_numbers, 2))
        keys = sorted([tuple(sorted(list(x))) for x in a])
        for key in keys:
            ret[key] = []

        for ipair in all_distances:
            key = tuple(sorted(atomic_number([self.structure.symbols[ipair[0]], self.structure.symbols[ipair[1]]])))
            if self.structure.is_periodic:
                ret[key] = np.concatenate((ret[key], all_distances[ipair]['distance']))
            else:
                ret[key].append(all_distances[ipair])

        # Sorting arrays
        for key in ret:
            ret[key].sort()
            ret[key] = np.array(ret[key])

        return ret

    def structure_distances(self, delta=0.01, sigma=0.01, integrated=True):
        dist_spec = self.all_distances_by_species()
        discrete_rdf = OrderedDict()
        nbins = int((self.radius + 5 * delta) / delta)
        discrete_rdf_x = np.arange(0, nbins * delta, delta)
        for spec_pair in dist_spec:
            discrete_rdf[spec_pair] = np.zeros(nbins)
            positive_distances = dist_spec[spec_pair][dist_spec[spec_pair] > 0]
            # log.debug('Pair %s' % str(spec_pair))
            for Rij in positive_distances:
                # Integrating for bin from - 8*sigma to +8*sigma centered on Rij
                # Values outside this range are negligible
                imin = int(max(0, (Rij - 8 * sigma) / delta))
                imax = int(min(len(discrete_rdf_x), (Rij + 8 * sigma) / delta))
                # log.debug('Computing for distance %7.3f for indices between %d and %d' % (Rij, imin, imax))
                for i in range(imin, imax):
                    x = discrete_rdf_x[i]
                    if not integrated:
                        discrete_rdf[spec_pair][i] += np.exp(-((x - Rij) ** 2) / (2 * sigma * sigma)) / (
                            4 * math.pi * Rij * Rij)
                    else:
                        discrete_rdf[spec_pair][i] += integral_gaussian(x, x + delta, Rij, sigma) / (
                            4 * math.pi * Rij * Rij)

        return discrete_rdf_x, discrete_rdf

    def fp_oganov(self, delta=0.01, sigma=0.01):
        struc_dist_x, struc_dist = self.structure_distances(delta=delta, sigma=sigma)
        fp_oganov = struc_dist.copy()
        vol = self.structure.volume
        for spec_pair in struc_dist:
            for i in range(len(struc_dist[spec_pair])):
                specie0 = atomic_symbol(spec_pair[0])
                specie1 = atomic_symbol(spec_pair[1])
                number_atoms0 = self.structure.composition[specie0]
                number_atoms1 = self.structure.composition[specie1]
                fp_oganov[spec_pair][i] *= vol / (delta * number_atoms0 * number_atoms1)
                fp_oganov[spec_pair][i] -= 1
        return struc_dist_x, fp_oganov

    def bonds_coordination(self, initial_cutoff_radius=0.8, use_laplacian=True, jump=0.01, tol=1E-15):

        cutoff_radius = initial_cutoff_radius
        ad = self.all_distances()
        bonds = {}
        while True:
            laplacian = np.zeros((self.structure.natom, self.structure.natom), dtype=np.int8)

            for pair in ad:
                atom1 = self.structure.symbols[pair[0]]
                atom2 = self.structure.symbols[pair[1]]
                sum_covalent_radius = sum(covalent_radius([atom1, atom2]))
                condition = np.bitwise_and(ad[pair]['distance'] < cutoff_radius * sum_covalent_radius,
                                           ad[pair]['distance'] > 0)
                bonds[pair] = ad[pair]['distance'][condition]
                if len(bonds[pair]) > 0:
                    laplacian[pair[0], pair[1]] = -1
                    laplacian[pair[1], pair[0]] = -1
            for i in range(self.structure.natom):
                laplacian[i, i] = 0
                laplacian[i, i] = -sum(laplacian[i])

            if use_laplacian:
                if np.max(np.abs(laplacian)) == 0:
                    cutoff_radius += jump

                ev = numpy.linalg.eigvalsh(laplacian)

                if sum(ev < tol) > 1:
                    cutoff_radius += jump
                else:
                    break
            else:
                break
        coordination = np.zeros(self.structure.natom, dtype=int)
        for pair in bonds:
            coordination[pair[0]] += len(bonds[pair])
            coordination[pair[1]] += len(bonds[pair])

        return bonds, coordination, round(cutoff_radius, 3)

    def get_bonds_coordination(self, initial_cutoff_radius=0.8, ensure_conectivity=False, use_laplacian=True,
                               verbose=False, tol=1E-15, jump=0.01, use_jump=True):
        """
        Computes simultaneously the bonds for all atoms and the coordination
        number using a multiplicative tolerance for the sum of covalent radius

        :param use_jump:
        :param jump:
        :param tol:
        :param verbose:
        :param use_laplacian:
        :param initial_cutoff_radius: (float) Tolerance factor (default is 1.2)
        :param ensure_conectivity: (bool) If True the tolerance of each bond is
               adjusted to ensure that each atom is connected at least once
        :return: tuple
        """
        if verbose:
            print('Computing all distances...')
        bonds_dict, distances_list = self.close_distances()
        if verbose:
            print('Number of distances computed: ', len(distances_list))

        cutoff_radius = initial_cutoff_radius
        bonds = None
        coordination = None
        tolerances = None

        while True:
            if verbose:
                print('Current cutoff radius : ', cutoff_radius)
            bonds = []
            tolerances = []
            for i in range(self.structure.natom):
                tole = cutoff_radius
                while True:
                    tmp_bonds = []
                    min_proportion = sys.float_info.max
                    for j in bonds_dict[str(i)]:
                        atom1 = self.structure.symbols[distances_list[j]['pair'][0]]
                        atom2 = self.structure.symbols[distances_list[j]['pair'][1]]
                        sum_covalent_radius = sum(covalent_radius([atom1, atom2]))
                        distance = distances_list[j]['distance']
                        if distance == 0.0:
                            continue
                        proportion = distance / sum_covalent_radius
                        min_proportion = min(min_proportion, proportion)
                        if proportion <= tole:
                            tmp_bonds.append(j)
                    if len(tmp_bonds) == 0 and ensure_conectivity:
                        tole = min_proportion
                        cutoff_radius = tole
                    else:
                        bonds.append(tmp_bonds)
                        tolerances.append(min_proportion)
                        break

            if use_laplacian:
                size = (self.structure.natom, self.structure.natom)
                laplacian = np.zeros(size, dtype=np.int8)
                for listibonds in bonds:
                    for ibond in listibonds:
                        data = distances_list[ibond]
                        i = data['pair'][0]
                        j = data['pair'][1]
                        laplacian[i, j] = -1
                        laplacian[j, i] = -1
                        # print '%d %d' % (i,j)
                for i in range(self.structure.natom):
                    laplacian[i, i] = 0
                    laplacian[i, i] = -sum(laplacian[i])
                if verbose:
                    print(laplacian)
                if np.max(np.abs(laplacian)) == 0:
                    cutoff_radius += jump
                    if verbose:
                        print('The laplacian is all zero')
                        print('Increasing cutoff radius by ', jump, 'A\n')
                    continue

                # if verbose:
                # print laplacian
                # evals, evecs = scipy.sparse.linalg.eigsh(laplacian)
                ev = numpy.linalg.eigvalsh(laplacian)
                if verbose:
                    print('Number of Eigenvalues close to zero :', sum(ev < tol))
                    print('Lowest Eigenvalues :', ev)
                    # print 'Lowest Eigenvalues :', evals

                if sum(ev < tol) > 1 and use_jump:
                    cutoff_radius += jump
                    if verbose:
                        print('Increasing cutoff radius by ', jump, 'A\n')
                else:
                    increase = False
                    for i in bonds:
                        if sum(i) == 0 and use_jump:
                            increase = True
                    if increase:
                        cutoff_radius += jump
                        if verbose:
                            print('Increasing cutoff radius by', jump, 'A\n')
                    else:
                        break
            else:
                break

        if bonds is not None:
            coordination = [len(x) for x in bonds]
        return bonds, coordination, distances_list, tolerances, cutoff_radius

    def hardness_XX(self, initial_cutoff_radius=0.8, use_laplacian=True):
        """
        Implementation of Hardness algorithm:
        
        First-principles structural design of superhard materials
        J. Chem. Phys. 138, 114101 (2013); https://doi.org/10.1063/1.4794424
        Xinxin Zhang, et al.
        """

        bonds, coordination, cutoff_radius = self.bonds_coordination(initial_cutoff_radius=initial_cutoff_radius,
                                                                     use_laplacian=use_laplacian)

        sigma = 3.0
        c_hard = 1300.0
        xprod = 1.
        tot = 0.0
        f_d = 0.0
        f_n = 1.0
        atomicnumbers = atomic_number(self.structure.species)

        pcm_log.debug('Atomic numbers in the structure : %s' % str(atomicnumbers))

        for i in atomicnumbers:
            f_d += valence(i) / covalent_radius(i)
            f_n *= valence(i) / covalent_radius(i)

        if f_d == 0:
            pcm_log.debug('Returning zero as hardness. f_d= %10.3f' % f_d)
            return 0.0
        f = 1.0 - (self.structure.nspecies * f_n ** (1.0 / self.structure.nspecies) / f_d) ** 2

        diff_bonds = [x for x in bonds if len(bonds[x]) > 0]

        for pair in diff_bonds:
            i1 = pair[0]
            i2 = pair[1]

            ei = valence(self.structure.symbols[i1]) / covalent_radius(self.structure.symbols[i1])
            ej = valence(self.structure.symbols[i2]) / covalent_radius(self.structure.symbols[i2])

            for dij in bonds[pair]:
                sij = math.sqrt(ei * ej) / (coordination[i1] * coordination[i2]) / dij
                xprod *= sij

            num_i_j_bonds = len(bonds[pair])
            pcm_log.debug('Number of bonds for pair %s = %d' % (str(pair), num_i_j_bonds))
            tot += num_i_j_bonds

        vol = self.structure.volume

        pcm_log.debug("Structure volume: %7.3f" % vol)
        pcm_log.debug("Total number of bonds: %d" % tot)
        pcm_log.debug("Bonds: %s" % str(bonds))

        hardness_value = (c_hard / vol) * tot * (xprod ** (1. / tot)) * math.exp(-sigma * f)

        return round(hardness_value, 3), cutoff_radius, coordination

    def hardness(self, verbose=True, initial_cutoff_radius=0.8, ensure_conectivity=False, use_laplacian=True,
                 use_jump=True, tol=1E-15):
        """
        Calculates the hardness of a structure based in the model of XX
        We use the covalent radii from pychemia.utils.periodic.
        If noupdate=False
        the Laplacian matrix method is not used and rcut is 2*max(cov_radii)

        :param use_jump:
        :param ensure_conectivity:
        :param verbose: (bool) To print some debug info
        :param initial_cutoff_radius: (float)
        :param use_laplacian: (bool) If True, the Laplacian method is used
        :param tol: (float) Tolerance for considering two atoms bonded

        :rtype : (float)
        """
        if self._supercell == (1, 1, 1) and verbose:
            print('''Only internal connectivity can be ensure, for complete connectivity in the crystal you must use a
                  supercell at of (2,2,2)''')

        bonds, coordination, all_distances, tolerances, cutoff_radius = \
            self.get_bonds_coordination(initial_cutoff_radius=initial_cutoff_radius,
                                        ensure_conectivity=ensure_conectivity,
                                        use_laplacian=use_laplacian, verbose=verbose, use_jump=use_jump, tol=tol)

        if verbose:
            print('Structure coordination : ', coordination)

        sigma = 3.0
        c_hard = 1300.0
        x = 1.
        tot = 0.0
        f_d = 0.0
        f_n = 1.0
        atomicnumbers = atomic_number(self.structure.species)

        if verbose:
            print('Atomic numbers in the structure :', atomicnumbers)

        for i in atomicnumbers:
            f_d += valence(i) / covalent_radius(i)
            f_n *= valence(i) / covalent_radius(i)

        # if verbose:
        # print 'fd', f_d
        #    print 'fn', f_n
        #    print atomicnumbers
        if f_d == 0:
            return 0.0
        f = 1.0 - (len(atomicnumbers) * f_n ** (1.0 / len(atomicnumbers)) / f_d) ** 2

        # Selection of different bonds
        diff_bonds = np.unique(np.array(functools.reduce(lambda xx, y: xx + y, bonds)))
        if verbose:
            print('Number of different bonds : ', len(diff_bonds))

        for i in diff_bonds:
            i1 = all_distances[i]['pair'][0]
            i2 = all_distances[i]['pair'][1]

            ei = valence(self.structure.symbols[i1]) / covalent_radius(self.structure.symbols[i1])
            ej = valence(self.structure.symbols[i2]) / covalent_radius(self.structure.symbols[i2])
            # print 'bond ->', sqrt(ei * ej), (coordination[i1] * coordination[i2]), all_distances[i]['distance']

            sij = math.sqrt(ei * ej) / (coordination[i1] * coordination[i2]) / all_distances[i]['distance']
            num_i_j_bonds = len([j for j in diff_bonds if i1 in all_distances[j]['pair'] and
                                 i2 in all_distances[j]['pair']])
            # print 'sij', sij
            # print 'num_i_j_bonds', num_i_j_bonds
            tot += num_i_j_bonds
            x *= sij
            # print 'x', x

        vol = self.structure.volume
        if verbose:
            print("Structure volume:", vol)
            # print("f:", f)
            # print("x:", x)

        # print 'len_bonds', len(diff_bonds
        hardness_value = c_hard / vol * (len(diff_bonds) * x ** (1. / (len(diff_bonds)))) * math.exp(-sigma * f)

        return round(hardness_value, 3), cutoff_radius, coordination

    def get_bonds(self, radius, noupdate=False, verbose=False, tolerance=0.05):
        """
        Calculates bond lengths between all atoms within "radius".

        :param radius: (float) The radius of the sphere were the bonds will be computed
        :param noupdate: (bool) If the original radius should be increased until a connected structure is obtained
        :param verbose: (bool) Print some info about the number of bonds computed
        :param tolerance: (float) Tolerance for creation of a bond

        :rtype : dict The values of the bonds are stored in self.info['bonds']

        """
        # The number of atoms in the cell
        n = self.structure.natom
        coordination = None
        dis_dic = None

        if verbose:
            print('Testing with %s of atoms in cell' % n)
            print('Starting with R_cut = %s' % radius)

        while True:
            size = (n, n)
            lap_m = np.zeros(size)
            dis_dic = {}
            ndifbonds = 0
            found = False
            # lap_mat[i,j] = lap_mat[j,i]
            for i in range(n - 1):
                for j in range(i + 1, n):
                    dis = self.structure.get_distance(i, j, with_periodicity=True)
                    if dis < radius:
                        if len(dis_dic) != 0:
                            for kstr, kj in dis_dic.items():
                                tstr = "%s%s" % (self.structure.symbols[i],
                                                 self.structure.symbols[j])
                                if abs(kj[0] - dis) < tolerance and tstr in kstr:
                                    dis_dic[kstr][1] += 1
                                    found = True
                                    break
                        if not found:
                            ndifbonds += 1
                            kstr = "%s%s%s%s" % (self.structure.symbols[i],
                                                 self.structure.symbols[j],
                                                 self.structure.symbols[i],
                                                 ndifbonds)
                            dis_dic[kstr] = [dis, 1, [i, j]]

                        lap_m[i][j] = -1
                        lap_m[j][i] = -1

            # lap_map[i,i]
            coordination = []
            for i in range(n):
                lap_m[i, i] = abs(sum(lap_m[i]))
                coordination.append(lap_m[i, i])

            # Get eigenvalues and check how many zeros;
            eigen = numpy.linalg.eig(lap_m)[0]

            nzeros = 0
            for i in eigen:
                if round(i.real, 5) == 0.0:
                    nzeros += 1
            if nzeros == 1 or (noupdate and len(dis_dic) > 0):
                break
            else:
                radius += 0.02

            if radius > 10.0:
                print("Cut off radius > 10.0")
                break

        if verbose:
            print('New cut off = %s' % radius)
            print('Bonds:')
            for i, j in dis_dic.items():
                print("  %s %s" % (i, j))

        return radius, coordination, dis_dic

    def hardness_old(self, noupdate=False, verbose=False, tolerance=0.05):
        """
        Calculates the hardness of a structure based in the model of XX
        We use the covalent radii from pychemia.utils.periodic.
        If noupdate=False
        the Laplacian matrix method is not used and rcut is 2*max(cov_radii)

        :param noupdate: (bool) If True, the Laplacian method is used
        :param verbose: (bool) To print some debug info
        :param tolerance: (float)

        :rtype : (float)
        """

        superc = self.structure.copy()
        superc.supercell(2, 2, 2)
        structure_analisys = StructureAnalysis(superc)

        natom = superc.natom
        volume = superc.volume

        max_covalent_radius = max(covalent_radius(superc.symbols))
        if verbose:
            print('Number of atoms', natom)
            print('Volume         ', volume)
            print('Covalent rad max', max_covalent_radius)
        rcut, coord, dis_dic = structure_analisys.get_bonds(2.0 * max_covalent_radius, noupdate, verbose, tolerance)

        sigma = 3.0
        c_hard = 1300.0
        x = 1.
        tot = 0.0
        f_d = 0.0
        f_n = 1.0
        dic_atms = {}
        for i in superc.symbols:
            dic_atms[i] = atomic_number(i)

        for i in dic_atms.keys():
            f_d += valence(i) / covalent_radius(i)
            f_n *= valence(i) / covalent_radius(i)
        f = 1.0 - (len(dic_atms) * f_n ** (1.0 / len(dic_atms)) / f_d) ** 2

        if verbose:
            print('BONDS')
            print(dis_dic)
            print('COORDINATION')
            print(coord)

        for i in dis_dic.keys():
            i1 = dis_dic[i][2][0]
            i2 = dis_dic[i][2][1]

            ei = valence(superc.symbols[i1]) / covalent_radius(superc.symbols[i1])
            ej = valence(superc.symbols[i2]) / covalent_radius(superc.symbols[i2])

            sij = math.sqrt(ei * ej) / (coord[i1] * coord[i2]) / dis_dic[i][0]

            tot += dis_dic[i][1]
            x *= sij * dis_dic[i][1]

        if verbose:
            print("V:", volume)
            print("f:", f)
            print("x:", x)

        hardness_value = c_hard / volume * (len(dis_dic) * x ** (1. / (len(dis_dic)))) * math.exp(-sigma * f)

        if verbose:
            print(hardness_value)

        return round(hardness_value, 3)



from abc import ABCMeta, abstractmethod
import numpy as np


class Relaxator:
    __metaclass__ = ABCMeta

    def __init__(self, target_forces):
        self.target_forces = target_forces

    @abstractmethod
    def run(self):
        pass

    @abstractmethod
    def get_forces_stress_energy(self):
        return None, None, None

    @abstractmethod
    def get_final_geometry(self):
        pass

    def get_max_force_stress(self):

        forces, stress, total_energy = self.get_forces_stress_energy()

        if forces is not None:
            max_force = np.max(np.apply_along_axis(np.linalg.norm, 1, forces))
        else:
            max_force = None

        if stress is not None:
            max_stress = np.max(np.abs(stress.flatten()))
        else:
            max_stress = None

        return max_force, max_stress

    def relaxation_status(self):

        max_force, max_stress = self.get_max_force_stress()

        if max_force is None:
            good_forces = False
        elif max_force < self.target_forces:
            good_forces = True
        else:
            good_forces = False

        if max_stress is None:
            good_stress = False
        elif max_stress < self.target_forces:
            good_stress = True
        else:
            good_stress = False

        return good_forces, good_stress



"""
The classes and methods of this package, take one or several structures and based only on geometrical
properties, computes fingerprints, modify structures, match atoms from two structures, split and create
surfaces.
The functionality provided by this package is critical for many of the operations done by global search
methods applied for structural search.
"""

from .analysis import StructureAnalysis
from .changer import StructureChanger
from .matching import StructureMatch
from .cluster import ClusterAnalysis, ClusterMatch
from .surface import rotate_along_indices
from . import splitting



#!D:\bld\pychemia_1592382025033\_h_env\python.exe

import pychemia
from pychemia.visual.searcher import *
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt


if len(sys.argv) != 3:
    print('Use:')
    print("Report_Metaheuristics.py dbname formula")
    exit(1)

gen_size = 16
name = sys.argv[1]
formula = sys.argv[2]
navg_energy = 6
letterpage = (11, 8.5)
host = "localhost"
user = "guilleaf"
passwd = "zxcvbnm"
ssl = True


if __name__ == '__main__':
    pdf = PdfPages(name + 'multipage.pdf')

    pcdb = pychemia.db.PyChemiaDB(name, host=host, user=user, passwd=passwd, ssl=ssl)
    popu = pychemia.population.RelaxStructures(pcdb)
    print(popu)

    tags = get_all_status(popu)
    for i in ['lock', 'target_forces', 'relaxation']:
        if i in tags:
            tags.remove(i)
    print(tags)

    colors = ['r', 'g', 'm', 'c', 'y', 'k']
    color_tags = {}
    for i in tags:
        if len(colors) == 0:
            colors = ['r', 'g', 'm', 'c', 'y', 'k']
        color_tags[i] = colors.pop(0)

    for itag in tags:
        pcdb = pychemia.db.PyChemiaDB(name, host=host, user=user, passwd=passwd, ssl=ssl)
        popu = pychemia.population.RelaxStructures(pcdb, tag=itag)
        patches = []

        plot_generation_chart(popu, gen_size)

    # Change of energy vs distance
    print('Plot: Change of energy vs distance')
    ret = {}
    color = {'replace_by_other': 'r', 'replace_by_random': 'b'}
    plt.figure(figsize=letterpage)
    for change in ['replace_by_other', 'replace_by_random']:
        print('Searching for: ', change)
        ret[change] = []
        for entry in popu.pcdb.db.generation_changes.find({'change': change}):
            id_from = entry['from']
            id_to = entry['to']
            # print "%s --> %s" % (id_from, id_to)
            x = popu.distance(id_from, id_to)
            if popu.is_evaluated(id_from) and popu.is_evaluated(id_to):
                values = popu.get_values([id_from, id_to])
                delta_energy = values[id_to] - values[id_from]
                y = delta_energy
                ret[change].append([x, y])
        ret[change] = np.array(ret[change])
        if len(ret[change]) > 0:
            plt.plot(ret[change][:, 0], ret[change][:, 1], color[change] + 'o', label=change)

    plt.xlim(min(ret[change][:, 0]) - 0.01, 0.01 + max(ret[change][:, 0]))
    plt.xlabel('Distance between structures (after relaxation)')
    plt.ylabel('Energy change')
    plt.title('Change of energy due to movement of structures')
    plt.legend()
    pdf.savefig()
    plt.close()

    # Getting all known Structures
    pcdb = pychemia.db.PyChemiaDB('ICSD_Pures', host='157.182.27.249', user=None, passwd=None, ssl=False)
    known_structures = {'energy_pa': [], 'spacegroup': []}
    number_known = 0
    for entry in pcdb.entries.find({'structure.formula': formula,
                                    'properties.energy_pa': {'$exists': 1}}, {'properties': 1}):
        known_structures['energy_pa'].append(entry['properties']['energy_pa'])
        known_structures['spacegroup'].append(entry['properties']['spacegroup'])
        number_known += 1

    # Best energies vs Generations
    miny = min(known_structures['energy_pa'])
    maxy = max(known_structures['energy_pa'])
    rangey = (miny - 0.1 * (maxy - miny), maxy + 0.1 * (maxy - miny))

    plt.figure(figsize=letterpage)

    maxx = 0
    for itag in tags:

        pcdb = pychemia.db.PyChemiaDB(name, host=host, user=user, passwd=passwd, ssl=ssl)
        popu = pychemia.population.RelaxStructures(pcdb, tag=itag)

        inigen = fingen = 0
        inigen, fingen = get_generation_limits(popu, gen_size)

        if fingen > maxx:
            maxx = fingen

        # All energies vs Generations
        sys.stdout.write("Plot: All energies vs Generations:")
        for j in range(inigen, fingen):
            sys.stdout.write(' %d' % j)
            thegen = get_generation(popu, j)
            energies = []
            for i in range(len(thegen)):
                struct, properties, status = popu.pcdb.get_dicts(thegen[i])
                energy = properties['energy_pa']
                energies.append(energy)
            if len(energies) > 0:
                for i in range(len(energies)):
                    plt.plot([j, j + 1 - 0.1], [energies[i], energies[i]], color_tags[itag] + '-')
                    plt.text(j + 0.9 * float(i) / len(energies), energies[i], str(i),
                             ha="center", family='sans-serif', size=6)
        sys.stdout.write('\n')

    for i in range(number_known):
        energy_pa = known_structures['energy_pa'][i]
        spacegroup = known_structures['spacegroup'][i]
        plt.plot([inigen, maxx], energy_pa * np.ones(2), 'b--')
        plt.text(fingen, energy_pa, str(spacegroup), ha="left", family='sans-serif', size=6, color='b')

    plt.xticks(range(maxx))
    plt.ylabel('Energy per atom [eV]')
    plt.xlabel('Generations')
    plt.title('Energies')
    pdf.savefig()
    plt.close()

    plt.figure(figsize=letterpage)
    maxx = 0
    for itag in tags:

        pcdb = pychemia.db.PyChemiaDB(name, host=host, user=user, passwd=passwd, ssl=ssl)
        popu = pychemia.population.RelaxStructures(pcdb, tag=itag)

        inigen, fingen = get_generation_limits(popu, gen_size)
        if fingen > maxx:
            maxx = fingen

        sys.stdout.write("Plot: Best energies vs Generations:")
        best = 16
        for j in range(inigen, fingen):
            sys.stdout.write(' %d' % j)
            thegen = get_generation(popu, j)[:best]
            energies = []
            spacegroups = []
            for i in range(len(thegen)):
                struct, properties, status = popu.pcdb.get_dicts(thegen[i])
                energies.append(properties['energy_pa'])
                spacegroups.append(properties['spacegroup'])
            if len(energies) > 0:
                for i in range(len(energies)):
                    if rangey[0] < energies[i] < rangey[1]:
                        plt.plot([j, j + 1 - 0.1], [energies[i], energies[i]], color_tags[itag] + '-')
                        plt.text(j + 0.9 * float(i) / len(energies), energies[i], str(i),
                                 ha="center", family='sans-serif', size=6, color='0.9')
                        plt.text(j + 0.9, energies[i], str(spacegroups[i]),
                                 ha="center", family='sans-serif', size=6, color=color_tags[itag])
        sys.stdout.write('\n')

    for i in range(number_known):
        energy_pa = known_structures['energy_pa'][i]
        spacegroup = known_structures['spacegroup'][i]
        plt.plot([inigen, maxx], energy_pa * np.ones(2), 'b--')
        plt.text(maxx, energy_pa, str(spacegroup), ha="left", family='sans-serif', size=6, color='b')

    plt.ylim(*rangey)
    plt.xticks(range(maxx))
    plt.ylabel('Energy per atom [eV]')
    plt.xlabel('Generations')
    plt.title('Energies')
    pdf.savefig()

    plt.figure(figsize=letterpage)

    for i in range(number_known):
        energy_pa = known_structures['energy_pa'][i]
        spacegroup = known_structures['spacegroup'][i]
        plt.plot([0, 1], energy_pa * np.ones(2), 'b--')
        plt.text(0.5, energy_pa, str(spacegroup), ha="left", family='sans-serif', size=6, color='b')

    index = 1
    for itag in tags:

        pcdb = pychemia.db.PyChemiaDB(name, host=host, user=user, passwd=passwd, ssl=ssl)
        popu = pychemia.population.RelaxStructures(pcdb, tag=itag)
        pre_energy = 1E10

        for entry_id in popu.ids_sorted(popu.evaluated):
            entry = popu.get_entry(entry_id)

            if itag in entry['status']:
                energy_pa = entry['properties']['energy_pa']
                spacegroup = entry['properties']['spacegroup']
                if np.abs(pre_energy - energy_pa) > 1E-3 and rangey[0] < energy_pa < rangey[1]:
                    plt.plot([index, index + 1], energy_pa * np.ones(2), color_tags[itag] + '--')
                    plt.text(index + 0.5, energy_pa, str(spacegroup),
                             ha="left",
                             family='sans-serif',
                             size=6,
                             color=color_tags[itag])
                    pre_energy = energy_pa

        index += 1

    plt.ylim(*rangey)
    plt.xticks(0.5 + np.arange(len(tags) + 1), ['CIFs'] + tags)
    plt.ylabel('Energy per atom [eV]')
    plt.xlabel('Structures found')
    plt.title('Energies')
    pdf.savefig()

    plt.figure(figsize=letterpage)

    for i in range(number_known):
        energy_pa = known_structures['energy_pa'][i]
        spacegroup = known_structures['spacegroup'][i]
        plt.plot([0, 1], energy_pa * np.ones(2), 'b-')
        # plt.text(0.5, energy_pa,str(spacegroup), ha="left", family='sans-serif', size=6, color='b')

    pcdb = pychemia.db.PyChemiaDB(name, host=host, user=user, passwd=passwd, ssl=ssl)
    popu = pychemia.population.RelaxStructures(pcdb, distance_tolerance=0.4, value_tol=0.1)
    pre_energy = 1E10

    selection = popu.evaluated
    n = len(selection)
    while True:
        selection = popu.cleaned_from_duplicates(selection)
        if len(selection) == n:
            break
        n = len(selection)
        print('Selection', n)

    for entry_id in popu.ids_sorted(selection):
        entry = popu.get_entry(entry_id)
        energy_pa = entry['properties']['energy_pa']
        spacegroup = entry['properties']['spacegroup']
        if np.abs(pre_energy - energy_pa) > 1E-3 and rangey[0] < energy_pa < rangey[1]:
            plt.plot([1, 2], energy_pa * np.ones(2), 'r-')
            # plt.text(1.5, energy_pa,str(spacegroup), ha="left", family='sans-serif', size=6, color='r')
            pre_energy = energy_pa

    plt.ylim(*rangey)
    plt.xticks(0.5 + np.arange(2), ['Known Structures', 'New predicted'])
    plt.ylabel('Energy per atom [eV]')
    plt.xlabel('Structures found')
    plt.title('Energies')
    pdf.savefig()

    pdf.close()



import os
import shutil
import json
import logging
from pychemia.code.vasp import VaspJob, VaspOutput, VaspInput
from pychemia.crystal import KPoints
from pychemia.code.vasp import read_poscar
from pychemia.utils.mathematics import round_small

__author__ = 'Guillermo Avendano-Franco'


class RelaxPopulation:
    def __init__(self, population, basedir, target_force=1E-2, target_stress=1E-2):
        self.population = population
        self.basedir = basedir
        self.vasp_jobs = {}
        self.runs = {}
        self.runner = None
        self.status = {}
        self.target_force = target_force
        self.target_stress = target_stress

    def create_dirs(self, clean=False):
        if not os.path.isdir(self.basedir):
            os.makedirs(self.basedir)
        elif clean:
            for i in os.listdir(self.basedir):
                shutil.rmtree(self.basedir + os.sep + i)
        for i in self.population.pcdb.entries.find():
            name = self.basedir + os.sep + str(i['_id'])
            if not os.path.isdir(name):
                os.mkdir(name)

    def create_inputs(self, kp_density=10000, encut=1.0):
        # kpoints = KPoints(kmode='gamma', grid=[4, 4, 4])
        for entry in self.population.pcdb.entries.find():
            name = str(entry['_id'])
            workdir = self.basedir + os.sep + name
            structure = self.population.db.get_structure(entry['_id'])
            kpoints = KPoints.optimized_grid(structure.lattice, kp_density=kp_density)
            print(kpoints)
            vj = VaspJob(workdir=workdir)
            vj.initialize(structure=structure, kpoints=kpoints)
            inp = VaspInput()
            inp.set_rough_relaxation()
            vj.set_input_variables(inp)
            vj.write_potcar()
            vj.input_variables.set_encut(ENCUT=encut, POTCAR=workdir + os.sep + 'POTCAR')
            vj.set_inputs()
            self.vasp_jobs[name] = vj
            self.runs[name] = 0
            self.status[name] = ['ACTIVE']

    def add_status(self, entry_id, value):
        if value not in self.status[entry_id]:
            self.status[entry_id].append(value)

    def del_status(self, entry_id, value):
        if value in self.status[entry_id]:
            self.status[entry_id].remove(value)

    def flip_status(self, entry_id, oldvalue, newvalue):
        self.del_status(entry_id, oldvalue)
        self.add_status(entry_id, newvalue)

    def modify_input(self, entry_id):
        if 'RELAXED' not in self.status[entry_id] and 'NOPROCAR' not in self.status[entry_id] \
                and 'NOOUTCAR' not in self.status[entry_id]:
            return True
        else:
            return False

    def update(self, workdir):
        """
        This routine determines how to proceed with the relaxation
        for one specific work directory

        :param workdir: (str) String representation of the id in the mongodb
        :return:
        """

        # workdir = self.basedir + os.sep + entry_id
        entry_id = os.path.basename(workdir)
        vj = self.vasp_jobs[entry_id]
        runj = self.runs[entry_id]
        if os.path.isfile(workdir + os.sep + 'OUTCAR'):
            vj.get_outputs()
        self.update_history(entry_id)

        if os.path.isfile(workdir + os.sep + 'RELAXED'):
            self.add_status(entry_id, 'RELAXED')
        elif not os.path.isfile(workdir + os.sep + 'PROCAR'):
            self.add_status(entry_id, 'NOPROCAR')
        else:
            self.del_status(entry_id, 'NOPROCAR')
            if not os.path.isfile(workdir + os.sep + 'OUTCAR'):
                self.add_status(entry_id, 'NOOUTCAR')
            else:
                self.del_status(entry_id, 'NOOUTCAR')
                print('-')
                vo = VaspOutput(workdir + os.sep + 'OUTCAR')
                relaxation_info = vo.relaxation_info()
                if len(relaxation_info) != 3:
                    print('[' + str(entry_id) + ']' + ' Missing some data in OUTCAR (forces or stress)')
                    self.add_status(entry_id, 'NOOUTCAR')

                print('[' + str(entry_id) + ']' + 'Results:')
                for i in relaxation_info:
                    print('[' + str(entry_id) + '] %20s %12.5e' % (i, relaxation_info[i]))

                # Conditions to consider the structure relaxed
                if relaxation_info['avg_force'] < self.target_force:
                    if relaxation_info['avg_stress_diag'] < self.target_stress:
                        if relaxation_info['avg_stress_non_diag'] < self.target_stress:
                            wf = open(workdir + os.sep + 'RELAXED', 'w')
                            for i in relaxation_info:
                                wf.write("%15s %12.3f" % (i, relaxation_info[i]))
                            wf.close()
                            wf = open(workdir + os.sep + 'COMPLETE', 'w')
                            for i in relaxation_info:
                                wf.write("%15s %12.3f" % (i, relaxation_info[i]))
                            wf.close()
                            self.add_status(entry_id, 'RELAXED')

        if self.modify_input(entry_id):

            # How to change ISIF
            if relaxation_info['avg_force'] < 0.1:
                if relaxation_info['avg_stress_diag'] < 0.1:
                    if relaxation_info['avg_stress_non_diag'] < 0.1:
                        vj.input_variables.variables['ISIF'] = 3
                    else:
                        vj.input_variables.variables['ISIF'] = 3
                else:
                    vj.input_variables.variables['ISIF'] = 3
            else:
                vj.input_variables.variables['ISIF'] = 2

            # How to change IBRION
            # if info['avg_force'] < 0.1 and info['avg_stress_diag'] < 0.1 and info['avg_stress_non_diag'] < 0.1:
            #    vj.input_variables.variables['IBRION'] = 1
            # elif info['avg_force'] < 1 and info['avg_stress_diag'] < 1 and info['avg_stress_non_diag'] < 1:
            #    vj.input_variables.variables['IBRION'] = 2
            # else:
            #    vj.input_variables.variables['IBRION'] = 3

            # How to change EDIFF
            if vj.input_variables.variables['EDIFF'] > 2 * 1E-4:
                vj.input_variables.variables['EDIFF'] = round_small(vj.input_variables.variables['EDIFF'] / 2)
            else:
                vj.input_variables.variables['EDIFF'] = 1E-4

            # How to change EDIFFG
            if vj.input_variables.variables['EDIFFG'] < - 2 * self.target_force:
                vj.input_variables.variables['EDIFFG'] = round_small(vj.input_variables.variables['EDIFFG'] / 2)
            else:
                vj.input_variables.variables['EDIFFG'] = - self.target_force

            # Print new values
            print('[' + str(entry_id) + ']' + 'New Values:')
            for i in ['ISIF', 'IBRION', 'EDIFF', 'EDIFFG']:
                print('[' + str(entry_id) + ']' + i + ' : ', vj.input_variables.variables[i])
            print('-')

            for i in ['OUTCAR']:
                if not os.path.exists(workdir + os.sep + i):
                    wf = open(workdir + os.sep + i, 'w')
                    wf.write('')
                    wf.close()
                log = logging.handlers.RotatingFileHandler(workdir + os.sep + i, maxBytes=1, backupCount=1000)
                log.doRollover()

            try:
                vj.structure = read_poscar(workdir + os.sep + 'CONTCAR')
            except ValueError:
                print('Error reading CONTCAR')

            vj.set_inputs()
            properties = vj.outcar
            status = self.status[entry_id]
            newentry = self.population.db.update(entry_id, structure=vj.structure, properties=properties, status=status)

            vj.save_json(workdir + os.sep + 'vaspjob.json')
            wf = open(workdir + os.sep + 'entry.json', 'w')
            json.dump(newentry, wf, sort_keys=True, indent=4, separators=(',', ': '))
            wf.close()
            return True
        else:
            vj.set_inputs()
            status = self.status[entry_id]
            newentry = self.population.db.update(entry_id, structure=vj.structure, status=status)

            vj.save_json(workdir + os.sep + 'vaspjob.json')
            wf = open(workdir + os.sep + 'entry.json', 'w')
            json.dump(newentry, wf, sort_keys=True, indent=4, separators=(',', ': '))
            wf.close()
            return True

    def update_history(self, entry_id):
        filename = 'pychemia_relaxation.json'
        filepath = self.basedir + os.sep + entry_id + os.sep + filename
        if not os.path.exists(filepath):
            wf = open(filepath, 'w')
            data = [self.vasp_jobs[entry_id].to_dict]
            json.dump(data, wf, sort_keys=True, indent=4, separators=(',', ': '))
            wf.close()
        else:
            rf = open(filepath, 'r')
            data = json.load(rf)
            rf.close()
            data.append(self.vasp_jobs[entry_id].to_dict)
            wf = open(filepath, 'w')
            json.dump(data, wf, sort_keys=True, indent=4, separators=(',', ': '))
            wf.close()

    @property
    def workdirs(self):
        return [self.basedir + os.sep + name for name in self.population.members]

    @property
    def active_workdirs(self):
        return [self.basedir + os.sep + name for name in self.population.actives]

    def run(self, runner):

        entries_ids = self.population.members

        def worker(workdir):
            wf = open(workdir + os.sep + 'LOCK', 'w')
            wf.write('')
            wf.close()
            runner.run()
            os.remove(workdir + os.sep + 'LOCK')

        def checker(workdir):
            if os.path.isfile(workdir + os.sep + 'LOCK'):
                return False
            return self.update(workdir)

        workdirs = [self.basedir + os.sep + i for i in self.population.actives]
        runner.run_multidirs(workdirs, worker, checker)

        if not self.is_running:
            self.run(runner)

    def set_run(self, code, runner, basedir, kp_density=10000, encut=1.1):

        self.runner = runner

        self.create_dirs(clean=True)
        self.create_inputs(kp_density=kp_density, encut=encut)



import os
import json
import subprocess
from abc import ABCMeta, abstractmethod


class Task:
    __metaclass__ = ABCMeta

    def __init__(self, structure, task_params=None, workdir='.', executable=None):
        self.structure = structure
        self.workdir = workdir
        if not os.path.isdir(self.workdir):
            os.mkdir(self.workdir)

        if task_params is not None:
            self.task_params = task_params
        else:
            self.task_params = {}
        self.executable = executable
        self.finished = False
        self.success = False
        self.started = False
        self.output = {}
        self.report_dir = self.workdir + os.sep + 'REPORT'

    @abstractmethod
    def run(self):
        pass

    @abstractmethod
    def plot(self):
        pass

    @abstractmethod
    def report(self):
        pass

    @abstractmethod
    def load(self):
        pass

    def save(self, filename=None):
        if filename is None:
            filename = self.workdir + os.sep + 'task.json'
        wf = open(filename, 'w')
        ret = {'task_params': self.task_params, 'output': self.output}
        json.dump(ret, wf, sort_keys=True, indent=4, separators=(',', ': '))
        wf.close()

    def status(self):
        if self.finished:
            print('Task finished')
        if self.started:
            print('Task started')
        if self.success:
            print('Task completed successfully')

    def report_end(self, html, file_format):
        from lxml import etree

        doctype = '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">'
        result = etree.tostring(html,
                                xml_declaration=True,
                                doctype=doctype,
                                encoding='utf-8',
                                standalone=False,
                                with_tail=False,
                                method='xml',
                                pretty_print=True)

        wf = open(self.report_dir + os.sep + 'index.html', 'w')
        wf.write(result)

        if file_format != 'html':
            cwd = os.getcwd()
            os.chdir(self.report_dir)
            stderr = open('pandoc_out.log', 'w')
            stdout = open('pandoc_err.log', 'w')
            sp = subprocess.Popen(['pandoc', 'index.html', '-o', 'visual.' + file_format], stderr=stderr, stdout=stdout)
            os.chdir(cwd)
            stderr.close()
            stdout.close()
            return sp



import re


def read_log(filename):

    rf = open(filename)
    data = rf.read()
    energy_data = re.findall('Total Energy\s[\-]+\s([\s\d\w/\-.=]*)Forces', data)

    energetics = {}
    if len(energy_data) > 0:
        for iline in energy_data[-1].split('\n'):
            if '=' in iline:
                energetics[iline.split('=')[0].strip()] = float(iline.split('=')[1])

    forces_data = re.findall('Total Forces:\s+\s([\s\d\w/\-+.=:()]*)\n \n', data)

    forces = []
    if len(forces) > 0:
        for iline in forces[-1].split('\n'):
            if '=' in iline:
                forces.append([float(x) for x in iline.split('=')[1].split()])

    ret = {'energetics': energetics, 'forces': forces}
    return ret


def write_inp(structure, filename):

    wf = open(filename, 'w')

    if structure.is_crystal:
        periodicity_tag = 0
    else:
        periodicity_tag = 1

    wf.write("%d %d\n" % (structure.natom, periodicity_tag))





#!D:\bld\pychemia_1592382025033\_h_env\python.exe

import os
import sys
import time
import logging
import itertools
import argparse
from multiprocessing import Pool, cpu_count, Process
import pychemia
from pychemia.utils.periodic import atomic_symbol

try:
    from qmpy import Entry
except ImportError:
    Entry = None
    print("Could not import 'qmpy' as needed to interface with the OQMD database")
    exit(1)


def run_one(a):
    """
    Take one OQMD 'Entry' object, search all the calculations associated and take the best calculation
    in order to insert its data into the PyChemia Database

    :param a: OQMD Entry object
    :return:
    """

    energy = 1E10
    best_calculation = None

    if a.calculation_set.count() > 0:

        if 'standard' in a.calculations:
            best_calculation = a.calculations['standard']
            calculation_name = 'standard'
        elif 'fine_relax' in a.calculations:
            best_calculation = a.calculations['fine_relax']
            calculation_name = 'fine_relax'
        elif 'coarse_relax' in a.calculations:
            best_calculation = a.calculations['coarse_relax']
            calculation_name = 'coarse_relax'
        elif 'static' in a.calculations:
            best_calculation = a.calculations['static']
            calculation_name = 'static'
        elif 'relaxation' in a.calculations:
            best_calculation = a.calculations['relaxation']
            calculation_name = 'relaxation'
        elif len(a.calculations) > 0:
            calculations = sorted(a.calculations.keys())
            print('Calculations found: %s, using the last one' % calculations)
            best_calculation = a.calculations[calculations[-1]]
            calculation_name = calculations[-1]
        else:
            print('ERROR: Count > 0 and no calculation found')

    if best_calculation is not None:
        structure_name = None
        if best_calculation.output is not None:
            structure_used = best_calculation.output
            structure_id = best_calculation.output_id
            from_output = True
        elif best_calculation.input is not None:
            print(
                'WARNING: No data was found from the output of the calculation, using input geometries and leaving '
                'energetics empty')
            structure_used = best_calculation.input
            structure_id = best_calculation.input_id
            from_output = False
    else:
        calculation_name = None
        if a.structures is not None and len(a.structures) > 0:
            struct_keys = sorted(a.structures.keys())
            print("WARNING: Calculation not found for %s. Structures found: %s using the first one " % (a, struct_keys))
            structure_used = a.structures[struct_keys[0]]
            structure_id = None
            from_output = False
            structure_name = struct_keys[0]
        else:
            print("ERROR: No calculation and no structure found for %s" % a)
            return None, None

    cell = structure_used.cell.T
    symbols = atomic_symbol(structure_used.atomic_numbers)
    reduced = structure_used.coords

    structure = pychemia.Structure(cell=cell, symbols=symbols, reduced=reduced)
    entry_id = a.id

    if best_calculation is not None:
        calculation_id = best_calculation.id
        energy_pa = best_calculation.energy_pa
        energy = best_calculation.energy
        band_gap = best_calculation.band_gap
        settings = best_calculation.settings

        try:
            spacegroup_number = best_calculation.output.spacegroup.number
        except AttributeError:
            spacegroup_number = None

    else:
        calculation_id = None
        energy_pa = None
        energy = None
        band_gap = None
        settings = None
        spacegroup_number = None
        from_output = False

    try:
        symm = pychemia.crystal.CrystalSymmetry(structure)
        sym2 = symm.number(1E-2)
    except ValueError:
        sym2 = None

    properties = {'oqmd': {'structure_id': structure_id,
                           'entry_id': entry_id,
                           'calculation_id': calculation_id,
                           'energy_pa': energy_pa,
                           'energy': energy,
                           'band_gap': band_gap,
                           'settings': settings,
                           'from_output': from_output,
                           'calculation_name': calculation_name,
                           'structure_name': structure_name,
                           'spacegroup_number': spacegroup_number},
                  'spacegroup_number': {'value': sym2, 'symprec': 1E-2}}

    return structure, properties


def getter(entry_ids, db_settings, current, start=0):
    pcdb = pychemia.db.get_database(db_settings)
    ret = []
    index = 0
    n = 0

    initial = start * jump
    final = min(start * jump + jump, len(entry_ids))
    print('Process: %2d Processing from %6d to %6d total: %d' % (start, initial, final, len(entry_ids)))

    for a_id in entry_ids[initial:final]:

        if a_id not in current[index:]:
            ret.append(a_id)
            n += 1
        else:
            index = current.index(a_id)
            # Removing duplicated entries
            if index + 1 < len(current) and current[index + 1] == a_id:
                print('We found at least one duplicate!')
                duplicate = False
                for entry in pcdb.db.pychemia_entries.find({'properties.oqmd.entry_id': a_id}):
                    if duplicate:
                        print('Removing PyChemiaDB entry: %s' % str(entry['_id']))
                        pcdb.db.pychemia_entries.remove({'_id': entry['_id']})
                    duplicate = True

    print('Process: %2d Entries missing: %3d' % (start, len(ret)))
    return ret


def setter(db_settings, to_insert):
    print('Processing %d entries - ' % len(to_insert), end='')
    pcdb = pychemia.db.get_database(db_settings)

    if hasattr(os, 'getppid'):  # only available on Unix
        print('parent process: %d - ' % os.getppid(), end='')
    print('process id: %d' % os.getpid())

    index = 0
    for oqmd_id in to_insert:
        if index % 2000 == 0:
            print(index, oqmd_id)
        index += 1
        structure = None
        properties = None

        a = Entry.objects.get(id=oqmd_id)
        structure, properties = run_one(a)

        if structure is not None:
            entry_id = '%d_%s_' % (structure.nspecies, structure.formula)
            n = len(entry_id)
            texto = '%0' + ('%d' % (28 - n)) + 'd'
            entry_id += texto % properties['oqmd']['entry_id']
            if n > 17:
                print("%2d - %s" % (28 - n, entry_id))
            pcdb.insert(structure, properties=properties, entry_id=entry_id)
    return 0


def getter_star(a_b):
    return getter(*a_b)


def setter_star(a_b):
    return setter(*a_b)


version = 0.1
jump = 10000

if __name__ == '__main__':

    parser = argparse.ArgumentParser(
        description='Create or Update a PyChemia Database from the OQMD Database (www.oqmd.org)')
    parser.add_argument('-dbname', metavar='<DATABASE>', type=str, help='Database Name', default='PyChemiaMasterDB')
    parser.add_argument('-port', metavar='<PORTNUMBER>', type=int, help='Port (default: 27017)', default=27017)
    parser.add_argument('-ssl', metavar='<SSL>', type=bool, help='Using SSL (default:no)', default=False)
    parser.add_argument('-user', metavar='<USERNAME>', type=str, help='Database Username', default=None)
    parser.add_argument('-host', metavar='<HOSTNAME>', type=str, help='Hostname (default: localhost)',
                        default='localhost')
    parser.add_argument('-nprocs', metavar='N', type=int,
                        help='Number of concurrent proccess (default: Number of CPUs)', default=None)

    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger('pychemia')
    logger.addHandler(logging.NullHandler())
    logger.setLevel(logging.INFO)

    db_settings = {'name': args.dbname, 'host': args.host, 'port': args.port, 'ssl': args.ssl}
    if args.user is not None:
        passwd = input('Password: ')
        db_settings['user'] = args.user
        db_settings['passwd'] = passwd

    print('Database settings: \n%s\b' % db_settings)
    pcdb = pychemia.db.get_database(db_settings)

    nitems = pcdb.entries.count()
    print('Number of entries in the current PyChemia Database: %d' % nitems)

    current = []
    for entry in pcdb.db.pychemia_entries.find({'properties.oqmd.entry_id': {'$exists': True}}):
        current.append(entry['properties']['oqmd']['entry_id'])
    current.sort()
    print('Number of entries coming from OQMD: %d' % len(current))

    print('Number of entries in OQMD...', end='')
    queryset = Entry.objects.all()
    entry_ids = [entry.id for entry in queryset]
    print('%d' % len(entry_ids))

    if args.nprocs is None:
        nprocs = cpu_count()
    else:
        nprocs = args.nprocs

    print('Creating a pool of %d processes for feeding the database' % nprocs)
    pool = Pool(processes=nprocs)

    argus = []
    a_args = range((len(entry_ids) / jump) + 1)

    to_insert = pool.map(getter_star, itertools.izip(itertools.repeat(entry_ids),
                                                     itertools.repeat(db_settings),
                                                     itertools.repeat(current), a_args), chunksize=1)

    pool.close()

    #    to_insert=to_insert[:20]
    print(len(to_insert))
    print(db_settings)

    ps = [None for x in range(nprocs)]
    counter = 0

    # QMPY does not support concurrent executions
    # while counter < len(to_insert):
    #     for i in range(nprocs):
    #         if ps[i] is None or not ps[i].is_alive():
    #             ps[i] = Process(target=setter, args=(db_settings,to_insert[counter]))
    #             ps[i].start()
    #             #ps[i].join()
    #             counter+=1
    #             if counter == len(to_insert):
    #                 break
    #             time.sleep(1)
    #     time.sleep(30)
    # print(ps)

    #    pool = Pool(processes=nprocs)
    #    ret = pool.map(setter_star, itertools.izip(itertools.repeat(db_settings), to_insert), chunksize=1)
    #    pool.close()

    for i in range(len(to_insert)):
        if len(to_insert[i]) > 0:
            setter(db_settings, to_insert[i])



import time
from multiprocessing import Pool, Process
import pychemia

__author__ = 'Guillermo Avendano-Franco'


def cluster_worker(db_settings):
    while True:
        pcdb = pychemia.db.get_database(db_settings)
        population = pychemia.population.LJCluster(pcdb)

        entry = population.pcdb.db.pychemia_entries.find_one({'status.' + population.tag: True,
                                                              'status.lock': {'$exists': False},
                                                              'properties': {}}, {'_id': 1})
        if entry is not None:
            population.pcdb.lock(entry['_id'])
            structure, properties, energy = population.evaluate(entry['_id'])
            population.pcdb.update(entry['_id'], structure=structure, properties=properties)
            population.pcdb.unlock(entry['_id'])
        else:
            break


def cluster_evaluator(db_settings, nparal):
    pcdb = pychemia.db.get_database(db_settings)
    population = pychemia.population.LJCluster(pcdb)
    population.recover()
    print('Staring evaluator for %s' % population.name)
    while True:
        entry = population.pcdb.db.pychemia_entries.find_one({'status.' + population.tag: True,
                                                              'status.lock': {'$exists': False},
                                                              'properties': {}}, {'_id': 1})

        if entry is None:
            time.sleep(2)
            create_pool = False
        else:
            create_pool = True

        if create_pool:
            pool = Pool(processes=nparal)
            pool.map(cluster_worker, nparal * [db_settings])
            pool.close()
            pool.join()


def cluster_launcher(db_settings, nparal):
    p = Process(target=cluster_evaluator, args=(db_settings, nparal))
    p.start()
    return p



"""
Set of classes and functions to manipulate

OCTOPUS 'inp' input files

"""

from .input import OctopusInput
from .output import OctopusOutput
from .run import OctopusRun
from . import analysis

# __all__ = filter(lambda s: not s.startswith('_'), dir())





"""
Routines used to interact between pymatgen objects and
pymatdis objects
"""
import numpy as np

try:
    import pymatgen.io.cif
except ImportError:
    raise ImportError

from pychemia import Structure
from pymatgen import Structure as PMG_Structure


def cif2structure(filename, primitive=False):
    """
    Read a given file  as a cif file, convert it into a
    pymatgen structure and finally into a pychemia
    structure

    :param filename: (str) Filename of CIF file that should be read
    :param primitive: (boolean) If primitive cell must be returned
    :return:
    """

    cifp = _pymatgen.io.cif.CifParser(filename)
    pmg_structs = cifp.get_structures(primitive)
    ret = []
    for i in pmg_structs:
        ret.append(pymatgen2pychemia(i))
    return ret


def pymatgen2pychemia(pmg_struct):
    """
    Converts a pymatgen structure object into a pychemia
    structure object

    :param pmg_struct: (pymatgen.Structure) Structure from pymatgen that will be converted
    :return:
    """

    # if not pmg_struct.is_ordered:
    #     return None

    cell = pmg_struct.lattice_vectors()
    reduced = np.array([])
    sites = []
    symbols = []
    occupancies = []
    index = 0
    for i in pmg_struct:
        for j in i.species_and_occu:
            symbols.append(j.symbol)
            occupancies.append(i.species_and_occu[j])
            sites.append(index)
        reduced = np.concatenate((reduced, i.frac_coords))
        index += 1

    reduced.reshape((-1, 3))
    return Structure(cell=cell, reduced=reduced, symbols=symbols, sites=sites, occupancies=occupancies)


def pychemia2pymatgen(structure):
    """
    Converts an pychemia structure into a pymatgen structure object

    :param structure: (pychemia.Structure) Structure to convert into pymatgen Structure object
    :return:
    """
    lattice = structure.cell
    coords = structure.reduced
    species = structure.symbols
    return PMG_Structure(lattice, species, coords)



from ..codes import CodeOutput
import os
import re
import json


class SiestaOutput(CodeOutput):

    def __init__(self, outputfile='siesta.out'):
        CodeOutput.__init__(self)
        self.outputfile = None
        self.output_values = None
        self.data = None
        if os.path.isfile(outputfile):
            self.outputfile = outputfile
        if self.is_finished:
            self.read()

    @property
    def is_finished(self):
        if self.outputfile is None:
            return False

        rf = open(self.outputfile)
        data = rf.read()
        rf.close()
        if data[-14:] == 'Job completed\n':
            return True
        else:
            return False

    def read(self):
        if not os.path.isfile(self.outputfile):
            raise ValueError("ERROR: Siesta outputfile not found: %s" % self.outputfile)
        rf = open(self.outputfile)
        self.data = rf.read()
        rf.close()

        subdata = re.findall("siesta: Final energy \(eV\):[\s\d\w\W]*\n\n", self.data)
        # print(subdata)
        if len(subdata) == 0:
            raise ValueError('No Final data could be retrieved')
        elif len(subdata) > 1:
            raise ValueError('ERROR: Wrong parsing of data')

        ret = {}
        for i in subdata[0].split('\n'):
            # Debugging parser
            # print('Line => %s' % i)
            if 'siesta:' in i:
                line = i.replace('siesta:', '').strip()
            else:
                line = i.strip()
            if 'Final energy' in line:
                master = line[:-1].strip()
            elif 'Atomic forces' in line:
                master = line[:-1].strip()
            elif 'Stress tensor' in line:
                master = line[:-1].strip()
            elif 'Cell volume' in line:
                ret['Cell volume'] = self.parse_line(line.split('=')[1])
            elif 'Pressure' in line:
                master = line[:-1].strip()
            elif '(Free)E+ p_basis*V_orbitals' in line:
                ret['(Free)E+ p_basis*V_orbitals'] = self.parse_line(line.split('=')[1])
            elif '(Free)Eharris+ p_basis*V_orbitals' in line:
                ret['(Free)Eharris+ p_basis*V_orbitals'] = self.parse_line(line.split('=')[1])
            elif 'Electric dipole (a.u.)' in line:
                ret['Electric dipole (a.u.)'] = self.parse_line(line.split('=')[1])
            elif 'Electric dipole (Debye)' in line:
                ret['Electric dipole (Debye)'] = self.parse_line(line.split('=')[1])
            elif 'Vacuum level (max, mean)' in line:
                ret['Vacuum level (max, mean)'] = self.parse_line(line.split('=')[1])
            elif 'Elapsed wall time (sec)' in line:
                ret['Elapsed wall time (sec)'] = self.parse_line(line.split('=')[1])
            elif line.strip() == '':
                continue
            elif line.strip()[:3] == '---' and line.strip()[-3:] == '---':
                continue
            elif "CPU execution times" in line:
                break
            elif master is not None:
                if '=' in line:
                    if master not in ret:
                        ret[master] = {}
                    key = line.split('=')[0].strip()
                    value = line.split('=')[1]
                    ret[master][key] = self.parse_line(value)
                else:
                    if master not in ret:
                        ret[master] = []
                    ret[master].append(self.parse_line(line))
        self.output_values = ret

    def parse_line(self, line):
        ret = []
        for i in line.split():
            try:
                value = int(i)
            except ValueError:
                try:
                    value = float(i)
                except ValueError:
                    value = i
            ret.append(value)
        if len(ret) == 1:
            ret = ret[0]
        return ret

    def show_parsed_data(self):
        if self.output_values is None:
            raise ValueError('No data has been parsed')
        else:
            print(json.dumps(self.output_values, sort_keys=True, indent=4, separators=(',', ': ')))



#!D:\bld\pychemia_1592382025033\_h_env\python.exe

import os
import sys
import pychemia.code.octopus


def helper():
    print(""" Set a variable in octopus
   Use:

       octopus_setvar.py --filename 'Octopus_Input_File' [ --set varname value ]...  [--del varname]...
   """)


if __name__ == '__main__':

    # Script starts from here
    if len(sys.argv) < 2:
        helper()
        sys.exit(1)

    filename = ''
    toset = {}
    todel = []
    for i in range(1, len(sys.argv)):
        if sys.argv[i].startswith('--'):
            option = sys.argv[i][2:]
            # fetch sys.argv[1] but without the first two characters
            if option == 'version':
                print('Version 1.0')
                sys.exit()
            elif option == 'help':
                helper()
                sys.exit()
            elif option == 'filename':
                filename = sys.argv[i + 1]
            elif option == 'set':
                toset[sys.argv[i + 1]] = sys.argv[i + 2]
            elif option == 'del':
                todel.append(sys.argv[i + 1])
            else:
                print('Unknown option. --' + option)

    # Set the variables
    if os.path.isfile(filename):
        data = pychemia.code.octopus.AbinitInput(filename)
        for i in toset.keys():
            data.variables[i] = toset[i]
        for i in todel:
            if i in data.variables.keys():
                data.variables.pop(i)
        data.write(filename)
    else:
        print('ERROR: no filename', filename)



from ..codes import CodeOutput


class ElkOutput(CodeOutput):

    @property
    def is_finished(self):
        return False

    def read(self):
        pass

    def __init__(self):
        CodeOutput.__init__(self)



"""
Routines related to stochastic optimization techniques
"""

from .annealing import SimulatedAnnealing
from .bee import BeeAlgorithm
from .firefly import FireFly
from .genetic import GeneticAlgorithm
from .grey import GreyWolf
from .harmony import HarmonySearch
from .swarm import ParticleSwarm
from .mhm import MinimaHoppingMethod
from .ant import AntColony

# __all__ = filter(lambda s: not s.startswith('_'), dir())



import numpy as np
from pychemia.utils.periodic import atomic_number, covalent_radius, cpk_colors


class StructurePovray:
    def __init__(self, structure):

        self.structure = structure
        self.distance = 10

    def create_pov(self):

        ret = """
#version 3.7;
#include "colors.inc"    // The include files contain
#include "stones.inc"    // pre-defined scene elements
#include "glass.inc"
background{rgb 0}

"""
        if self.structure.is_crystal:
            self.distance = max(self.structure.lattice.lengths)
        else:
            self.distance = 10

        ret += "#declare r=%7.3f;\n #declare s=%7.3f;" % (self.distance, self.distance)

        ret += "camera {\n"
        ret += "\tlocation <%7.3f, %7.3f, %7.3f>\n" % (1.3 * self.distance, 1.3 * self.distance, -1.3 * self.distance)
        ret += "\tlook_at  <%7.3f, %7.3f, %7.3f>\n" % tuple(0.5 * sum(self.structure.cell[:]))
        ret += "}\n\n"

        if self.structure.nsites > 0:
            d = self.distance
            ret += "light_source { <%7.3f, %7.3f, %7.3f> color White}\n" % (2 * d, 2 * d, 2 * d)

        for imagx in np.arange(-1, 2):
            for imagy in np.arange(-1, 2):
                for imagz in np.arange(-1, 2):

                    for site in self.structure:
                        for symbol in site.symbols:
                            cell = self.structure.cell
                            x = site.position[0] - imagx * cell[0, 0] - imagy * cell[1, 0] - imagz * cell[2, 0]
                            y = site.position[1] - imagx * cell[0, 1] - imagy * cell[1, 1] - imagz * cell[2, 1]
                            z = site.position[2] - imagx * cell[0, 2] - imagy * cell[1, 2] - imagz * cell[2, 2]
                            if (x - self.distance) ** 2 + (y - self.distance) ** 2 + (z + self.distance) ** 2 < 2:
                                continue
                            cr = 0.5 * covalent_radius(symbol)
                            rgb = cpk_colors[atomic_number(symbol)]
                            color = 'rgb < %7.3f, %7.3f, %7.3f>' % (rgb[0], rgb[1], rgb[2])
                            ret += "sphere {\n"
                            ret += "\t<%7.3f, %7.3f, %7.3f>, %7.3f\n\ttexture {\n" % (x, y, z, cr)
                            ret += "\t\tpigment { color %s filter 0.4 transmit %7.3f}\n" % \
                                   (color, 1 - 0.9 * np.exp(-0.1 * (abs(imagx) + abs(imagy) + abs(imagz))))
                            ret += "\t\tnormal { bumps 0.8 scale 0.1 }\n\t\tfinish { phong %7.3f }\n\t}\n}\n\n" % \
                                   np.exp(-0.1 * (abs(imagx) + abs(imagy) + abs(imagz)))

                            if self.structure.nsites <= 0:
                                ret += "light_source { <%7.3f, %7.3f, %7.3f> color White}\n" % (x, y, z)

        ret += """union{
#include "cell.pov"
    scale 1
    rotate <0, 0, 0>
    pigment{rgb <0.3,0.3,0.9>} finish{phong 0.9 ambient 0.42 reflection 0.1}
}
"""
        return ret

    def write_povray(self, filename):

        wf = open(filename, 'w')
        wf.write(self.create_pov())
        wf.close()

        if self.structure.is_crystal:
            self.write_cell('cell.pov')

    def write_cell(self, filename):
        wf = open(filename, 'w')
        ret = ''
        for i in range(3):
            for j in range(3):
                ret += "cylinder { "
                if i == j:
                    ret += " <%7.3f, %7.3f, %7.3f>, " % (0.0, 0.0, 0.0)
                    ret += " <%7.3f, %7.3f, %7.3f>, " % tuple(self.structure.cell[j])
                else:
                    ret += " <%7.3f, %7.3f, %7.3f>, " % tuple(self.structure.cell[i])
                    ret += " <%7.3f, %7.3f, %7.3f>, " % tuple(self.structure.cell[i] + self.structure.cell[j])

                ret += " %7.3f }\n" % (self.distance / 100.0)

            ret += "cylinder { "
            ret += " <%7.3f, %7.3f, %7.3f>, " % tuple(sum(self.structure.cell[:]))
            ret += " <%7.3f, %7.3f, %7.3f>, " % tuple(sum(self.structure.cell[:]) - self.structure.cell[i])

            ret += " %7.3f }\n" % (self.distance / 100.0)
        wf.write(ret)
        wf.close()

# ret += "\topen           // Remove end caps\n"
#                #ret += "\ttexture { %s }\n" % ('T_Stone25 scale 4')
#                ret += "\ttexture { %s }\n" % ('pigment { Col_Glass_Old }')




import os
import numpy as np
from pychemia.utils.periodic import atomic_symbol, covalent_radius, atomic_number
from pychemia.utils.constants import bohr_angstrom, angstrom_bohr
from pychemia.utils.mathematics import unit_vectors
from .parser import parser
from pychemia.utils.netcdf import netcdf2dict
from pychemia.core import Structure
from ..codes import CodeInput

"""
Definition of the class input to read
ABINIT input files and store their information
as a python dictionary called 'variables'
"""

__author__ = "Guillermo Avendano-Franco"
__copyright__ = "Copyright 2016"
__version__ = "1.1"
__maintainer__ = "Guillermo Avendano-Franco"
__email__ = "guillermo.avendano@uclouvain.be"
__status__ = "Development"
__date__ = "May 13, 2016"


class AbinitInput(CodeInput):

    def __init__(self, input_filename=None):
        CodeInput.__init__(self)
        if input_filename is None:
            self.input_file = 'abinit.in'
        else:
            self.input_file = input_filename
        if os.path.isfile(self.input_file):
            if self.input_file[-6:] == 'OUT.nc':
                self.variables = netcdf2dict(self.input_file)
            else:
                self.read()

    def __import_input(self, filename):
        """
        Read an ABINIT input file and return a python dictionary
        with the input variables readed from that. The keys are
        the fullname input variables (acell,xcart21,etc). The
        values are numbers or lists except for
        the value '*[NUMBER]' that is keeped as string, and
        the string associated to the variable xyzfile

        Args:
            filename:
                ABINIT input filename
        """
        ans = parser(filename)
        if ans is not None:
            self.variables = ans

    def read(self):
        if not os.path.isfile(self.input_file):
            raise ValueError("ERROR: Could not read %s" % self.input_file)

        ans = parser(self.input_file)
        if ans is not None:
            self.variables = ans

    def check(self):
        if self.get_number_variables > 0:
            print("ABINIT input is readable and has %d variables" % self.get_number_variables)

    def __str__(self):
        """
        String representation of the object
        """
        ret = ''
        thekeys = self.variables.keys()
        varnames = [x for x in thekeys if not x[-1].isdigit()]

        if 'ndtset' in varnames:
            ret = ret + "#" + 60 * "-" + "\n#" + " MULTI DATASET\n#" + 60 * "-" + "\n\n"
            ret += self.write_key('ndtset')
            varnames.remove('ndtset')
            if 'jdtset' in varnames:
                ret += self.write_key('jdtset')
                varnames.remove('jdtset')
            if 'udtset' in varnames:
                ret += self.write_key('udtset')
                varnames.remove('udtset')
            ret += '\n'

        seqvarnames = [x for x in varnames if
                       (x[-1] == ':' or x[-1] == "+" or x[-1] == "?" or x[-2] == ':' or x[-2] == "+" or x[
                           -2] == "?")]
        if len(seqvarnames) > 0:
            ret = ret + "#" + 60 * "-" + "\n#" + " SEQUENCE\n#" + 60 * "-" + "\n\n"
            seqvarnames.sort()
            for i in seqvarnames:
                ret += self.write_key(i)
                varnames.remove(i)
            ret += '\n'

        if len(varnames) > 0:
            varnames.sort()
            ret = ret + "#" + 60 * "-" + "\n#" + " ALL DATASETS\n#" + 60 * "-" + "\n\n"
            for i in varnames:
                if i == 'dmatpawu' and 'lpawu' in self.variables:
                    if 2 in self.variables['lpawu']:
                        ret += self.write_key(i, ncolumns=5)
                    elif 3 in self.variables['lpawu']:
                        ret += self.write_key(i, ncolumns=7)
                else:
                    ret += self.write_key(i)
            ret += '\n'

        for dtset in range(1, 100):
            varnames = [x for x in thekeys if
                        (x[-len(str(dtset)):] == str(dtset) and not x[-len(str(dtset)) - 1:].isdigit())]
            if len(varnames) > 0:
                varnames.sort()
                ret = ret + "#" + 60 * "-" + "\n#" + " DATASET " + str(dtset) + "\n#" + 60 * "-" + "\n\n"
                for i in varnames:
                    if i == 'dmatpawu' and 'lpawu' in self.variables:
                        if 2 in self.variables['lpawu']:
                            ret += self.write_key(i, ncolumns=5)
                        elif 3 in self.variables['lpawu']:
                            ret += self.write_key(i, ncolumns=7)
                    ret += self.write_key(i)
                ret += '\n'
        return ret

    def write_key(self, varname, ncolumns=None, debug=False):
        """
        Receives an input variable and write their contents
        properly according with their kind and length

        Args:
            varname:
                The name of the input variable
            ncolumns:
                Number of columns for the input variable
            debug:
                Shows contents of variable before creating string out of it
        """

        ret = ''
        if varname not in self.variables:
            print("[ERROR] input variable: '%s' is not defined" % varname)
            return

        # Assume that the variables are integer and test if such assumption
        # is true
        integer = True
        real = False
        string = False
        compact = True

        if isinstance(self.variables[varname], (int, float)):
            varlist = [self.variables[varname]]
        elif isinstance(self.variables[varname], str):
            varlist = [self.variables[varname]]
        else:
            varlist = self.variables[varname]

        if debug:
            print('varlist: %s' % varlist)
            
        # Get the general kind of values for the input variable
        for j in varlist:
            try:
                if not float(j).is_integer():
                    # This is the case of non integer values
                    integer = False
                    real = True
                    string = False
                    if len(str(float(j))) > 7:
                        compact = False

            except ValueError:
                # This is the case of '*1' that could not
                # be converted because we don't know the size
                # of the array
                integer = False
                real = False
                string = True

        ret = ret + (varname.rjust(15)) + "  "

        known_variables = {'xred': [3], 'acell': [3]}

        if varname in known_variables:
            for i in known_variables[varname]:
                if len(varlist) % i == 0:
                    for j in range(int(len(varlist) / i)):
                        if j == 0:
                            ret += (i * '%17.10E ' + '\n') % tuple(varlist[j * i:j * i + i])
                        else:
                            ret += (17 * ' ' + i * '%17.10E ' + '\n') % tuple(varlist[j * i:j * i + i])
        elif ncolumns is not None:
            i = ncolumns
            for j in range(int(len(varlist) / i)):
                if j == 0:
                    ret += (i * '%17.10E ' + '\n') % tuple(varlist[j * i:j * i + i])
                else:
                    ret += (17 * ' ' + i * '%17.10E ' + '\n') % tuple(varlist[j * i:j * i + i])
        else:
            if debug:
                print("real: %s  integer: %s  string: %s" % (real, integer, string))

            for j in range(len(varlist)):

                if real:
                    if compact:
                        ret += ("%g" % varlist[j]).rjust(8)
                    else:
                        ret += "%17.10e" % varlist[j]
                elif integer:
                    ret += "%d" % varlist[j]
                elif string:
                    ret += "%s" % varlist[j]

                # Conditions to jump to a new line
                if ((j + 1) % 3) == 0 and real and j < len(varlist) - 1:
                    ret += "\n"
                    ret += 17 * " "
                elif j < len(varlist) - 1:
                    ret += " "
            ret += "\n"
        return ret

    def get_structure(self, idtset=None, units='bohr'):
        """
        Return the atomic structure from the input object
        for a given dataset (no dataset by default)
        """
        # NATOM
        natom = self.get_value('natom', idtset)
        # SYMBOLS
        ntypat = self.get_value('ntypat', idtset)
        symbols = []
        znucl = self.get_value('znucl', idtset)
        typat = self.get_value('typat', idtset)
        for i in range(natom):
            # NOTE: znucl is a real number in OUT.nc
            # Alchemic mixing is not allow here
            if ntypat == 1:
                symbols.append(atomic_symbol(int(znucl)))
            else:
                symbols.append(atomic_symbol(int(znucl[typat[i] - 1])))

        # POSITIONS
        xangst = self.get_value('xangst', idtset)
        xcart = self.get_value('xcart', idtset)
        xred = self.get_value('xred', idtset)

        rprim = self.get_value('rprim', idtset)
        acell = np.array(self.get_value('acell', idtset))

        # Set rprimd and acell using the default values
        # if not found
        if rprim is None:
            rprim = np.identity(3)
        else:
            rprim = np.array(rprim).reshape((3, 3))
        if acell is None:
            acell = np.ones(3)

        rprimd = np.zeros((3, 3))
        rprimd[0] = rprim[0] * acell
        rprimd[1] = rprim[1] * acell
        rprimd[2] = rprim[2] * acell

        if xangst is not None:
            xangst = np.array(xangst)
            positions = xangst.reshape((natom, 3))
            if units == 'bohr':
                positions = positions * angstrom_bohr
        elif xcart is not None:
            xcart = np.array(xcart)
            positions = xcart.reshape((natom, 3))
            if units == 'angstrom':
                positions = positions * bohr_angstrom
        elif xred is not None:
            xred = np.array(xred)
            xred = xred.reshape((natom, 3))
            xcart = np.zeros((natom, 3))

            # print rprimd
            # print xred

            for i in range(natom):
                xcart[i] = xred[i, 0] * rprimd[0] + xred[i, 1] * rprimd[1] + xred[i, 2] * rprimd[2]

            positions = xcart
            if units == 'angstrom':
                positions = positions * bohr_angstrom

        else:
            positions = None

        if units == 'angstrom':
            rprimd = rprimd * bohr_angstrom

        # Create an object atomic_structure
        structure = Structure(natom=natom, symbols=symbols, positions=positions, cell=rprimd)

        return structure

    def from_structure(self, structure):
        """
        Set input variables for a given structure

        :param structure: (pychemia.Structure) Structure to set ABINIT input variables
        :return:
        """
        natom = structure.natom
        ntypat = len(structure.species)
        znucl = atomic_number(structure.species)
        typat_dict = {}
        index = 1
        for ispec in structure.species:
            typat_dict[ispec] = index
            index += 1
        typat = [typat_dict[i] for i in structure.symbols]
        xcart = angstrom_bohr * structure.positions.flatten()
        acell = angstrom_bohr * np.array(structure.lattice.lengths)
        rprim = unit_vectors(structure.cell).T.flatten()
        for i in ['natom', 'ntypat', 'znucl', 'typat', 'xcart', 'acell', 'rprim']:
            self.set_value(i, eval(i))

    def get_dtsets_keys(self):
        """
        Get the list of dtset suffix acording to
        the values given by ndtset, jdtset and udtset
        """
        ret = None
        if 'jdtset' in self.variables and 'udtset' in self.variables:
            print('ERROR: Both udtset and jdtset could not be used ')
            return None

        elif 'ndtset' in self.variables:
            ndtset = self.get_value('ndtset')

            if ndtset != 0:
                if 'jdtset' in self.variables:
                    ret = list(self.variables['jdtset'])
                elif 'udtset' in self.variables:
                    ret = []
                    udtset = self.get_value('udtset')
                    for i in range(1, udtset[0] + 1):
                        for j in range(1, udtset[1] + 1):
                            print(ret)
                            print(str(i) + str(j))
                            ret.append(str(i) + str(j))
                else:
                    ret = range(1, ndtset + 1)
        else:
            ret = ['']

        return ret

    def atom_name(self, iatom, idtset=''):
        """
        Return the name of the atom and the position in the
        list of atoms such as H3, N4, etc
        """
        atomnumber = self.get_value('znucl', idtset=idtset)
        print("atomnumber=%s" % atomnumber)
        if isinstance(atomnumber, list):
            atomnumber = atomnumber[self.get_value('typat', idtset=idtset)[iatom] - 1]
        return atomic_symbol(atomnumber) + str(iatom + 1)

    def view_projections(self):
        """
        Show the 3 projections of the molecule in a single
        figure
        """
        import matplotlib.patches as mpatches
        from matplotlib.collections import PatchCollection
        from matplotlib.pylab import subplots

        fig, ax = subplots(nrows=1, ncols=3)
        fig.set_size_inches(15, 4)
        color = ['r', 'g', 'b']
        j = 0

        structure = self.get_structure()

        for i in structure.cell:
            ax[0].plot([0, i[0]], [0, i[1]], color[j] + '-', lw=3)
            ax[1].plot([0, i[1]], [0, i[2]], color[j] + '-', lw=3)
            ax[2].plot([0, i[2]], [0, i[0]], color[j] + '-', lw=3)
            j += 1

        proj = [[0, 1], [1, 2], [2, 0]]
        labels = [['x', 'y'], ['y', 'z'], ['z', 'x']]

        for j in range(3):

            patches = []
            for i in range(structure.natom):
                radius = 0.5 * covalent_radius(atomic_number(structure.symbols[i]))
                pos = structure.positions[i]
                art = mpatches.Circle((pos[proj[j][0]], pos[proj[j][1]]), radius, fc='g', ec='g')
                patches.append(art)

            collection = PatchCollection(patches, color='k', alpha=0.5)

            col = ax[j].add_collection(collection)
            ax[j].set_xlim(min(structure.positions[:, proj[j][0]]) - 1, max(structure.positions[:, proj[j][0]]) + 1)
            ax[j].set_ylim(min(structure.positions[:, proj[j][1]]) - 1, max(structure.positions[:, proj[j][1]]) + 1)
            ax[j].set_aspect('equal', adjustable='datalim')
            ax[j].set_xlabel(labels[j][0])
            ax[j].set_ylabel(labels[j][1])

        return fig, ax

    def clean(self):
        self.variables = {}

    def has_variable(self, varname, section=None):
        return varname in self.variables

    def get_value(self, varname, idtset=None, return_iterable=False):
        """
        Get the value of the input variable 'varname'
        associated with the dataset 'idtset'
        If 'idtset' is not given will asume that the
        value is not dataset dependent
        """
        name = ''
        fact = 1
        delta = 0

        # Get the right key for the abinit variable
        if idtset is None:
            if varname in self.variables:
                name = varname
        else:
            if (varname + str(idtset)) in self.variables:
                name = varname + str(idtset)
            elif idtset > 10:
                if (varname + '?' + (str(idtset)[1])) in self.variables:
                    name = varname + '?' + (str(idtset)[1])
                elif (varname + (str(idtset)[0]) + '?') in self.variables:
                    name = varname + (str(idtset)[0]) + '?'
                elif (varname + '+?') in self.variables and (varname + ':?') in self.variables:
                    name = varname + ':?'
                    fact = int(str(idtset)[0]) - 1
                    delta = self.variables[varname + '+?']
                elif (varname + '?+') in self.variables and (varname + '?:') in self.variables:
                    name = varname + '?:'
                    fact = int(str(idtset)[1]) - 1
                    delta = self.variables[varname + '?+']
            if name == '' and varname in self.variables:
                name = varname

        # print 'varname=',varname,'name=',name
        #  Get the value of the abinit variable
        if name != '':
            if isinstance(self.variables[name], list):
                npvalue = list(np.array(self.variables[name]) + fact * np.array(delta))
            else:
                npvalue = self.variables[name] + fact * delta

        elif (varname + ":") in self.variables and (varname + "+") in self.variables:
            if isinstance(self.variables[varname + ":"], list):
                npvalue = list(np.array(self.variables[varname + ":"]) +
                               (idtset - 1) * np.array(self.variables[varname + "+"]))
            else:
                npvalue = self.variables[varname + ":"] + (idtset - 1) * self.variables[varname + "+"]
        else:
            npvalue = None

        if isinstance(npvalue, (int, float)) and return_iterable:
            npvalue = [npvalue]

        return npvalue

    def set_value(self, varname, value, idtset=''):
        """
        Set the value 'value' into the dictionary
        input with key 'varname'+str(idtset)
        The value could be an integer, real, list
        or numpy array, the internal representation
        is always serializable.
        """
        if isinstance(value, (int, float)):
            npvalue = value
        elif isinstance(value, np.ndarray):
            if value[0].dtype == np.dtype('>f8'):
                npvalue = [round(x, 11) for x in value.flatten()]
            elif value[0].dtype == np.dtype('>i4'):
                npvalue = [int(x) for x in value.flatten()]
            else:
                npvalue = list(value)
        else:
            npvalue = list(value)

        if idtset == '':
            self.variables[varname] = npvalue
        else:
            self.variables[varname + str(idtset)] = npvalue


# class InputVariables(collections.MutableMapping):
#     """
#     An input object contains:

#     data:
#           variables = Dictionary whose keys are ABINIT variable names

#     methods:
#             write = Write the input into as a text file that ABINIT
#                     can use as an input file

#             get_value = Get the value of a particular variable
#             set_value = Set the value of a particular variable

#             get_atomic_structure =
#     """


#     def __init__(self, *args, **kwargs):
#         """
#         Creates a new input object, the input object
#         contains a dictionary whose keys are the abinit
#         input variables and the values are always serializable
#         """
#         self.variables = {}
#         filename = ''
#         if len(args) == 1:
#             x = args[0]
#             if isinstance(x, AbiFiles):
#                 filename = x.basedir + "/" + x.files["in"]
#             elif os.path.isfile(x):
#                 filename = x

#         if 'filename' in kwargs:
#             filename = kwargs['filename']

#         if os.path.isfile(filename):
#             if filename[-3:] == '.in':
#                 self.__import_input(filename)
#             elif filename[-6:] == '.files':
#                 abifile = AbiFiles(filename)
#                 filename = abifile.get_input_filename()
#                 self.__import_input(filename)
#             elif filename[-6:] == 'OUT.nc':
#                 self.variables = netcdf2dict(filename)
#             else:
#                 try:
#                     self.__import_input(filename)
#                 except ValueError:
#                     print('File format not identified')

#     def __delitem__(self, key):
#         return self.variables.__delitem__(key)

#     def __setitem__(self, key, value):
#         return self.variables.__setitem__(key, value)

#     def __getitem__(self, key):
#         return self.variables.__getitem__(key)

#     def __iter__(self):
#         return self.variables.__iter__()

#     def __len__(self):
#         return self.variables.__len__()

#     def __import_input(self, filename):
#         """
#         Read an ABINIT input file and return a python dictionary
#         with the input variables readed from that. The keys are
#         the fullname input variables (acell,xcart21,etc). The
#         values are numbers or lists except for
#         the value '*[NUMBER]' that is keeped as string, and
#         the string associated to the variable xyzfile

#         Args:
#             filename:
#                 ABINIT input filename
#         """
#         ans = parser(filename)
#         if ans is not None:
#             self.variables = ans

#     def write(self, filename):
#         """
#         Write an input object into a text
#         file that ABINIT can use as an input
#         file

#         Args:
#             filename:
#                 The 'abinit.in' filename that will be written
#         """
#         wf = open(filename, 'w')
#         wf.write(self.__str__())
#         wf.close()

#     def __str__(self):
#         """
#         String representation of the object
#         """
#         ret = ''
#         thekeys = self.variables.keys()
#         varnames = [x for x in thekeys if not x[-1].isdigit()]

#         if 'ndtset' in varnames:
#             ret = ret + "#" + 60 * "-" + "\n#" + " MULTI DATASET\n#" + 60 * "-" + "\n\n"
#             ret += self.write_key('ndtset')
#             varnames.remove('ndtset')
#             if 'jdtset' in varnames:
#                 ret += self.write_key('jdtset')
#                 varnames.remove('jdtset')
#             if 'udtset' in varnames:
#                 ret += self.write_key('udtset')
#                 varnames.remove('udtset')
#             ret += '\n'

#         seqvarnames = [x for x in varnames if
#                        (x[-1] == ':' or x[-1] == "+" or x[-1] == "?" or x[-2] == ':' or x[-2] == "+" or x[-2] == "?")]
#         if len(seqvarnames) > 0:
#             ret = ret + "#" + 60 * "-" + "\n#" + " SEQUENCE\n#" + 60 * "-" + "\n\n"
#             seqvarnames.sort()
#             for i in seqvarnames:
#                 ret += self.write_key(i)
#                 varnames.remove(i)
#             ret += '\n'

#         if len(varnames) > 0:
#             varnames.sort()
#             ret = ret + "#" + 60 * "-" + "\n#" + " ALL DATASETS\n#" + 60 * "-" + "\n\n"
#             for i in varnames:
#                 ret += self.write_key(i)
#             ret += '\n'

#         for dtset in range(1, 100):
#             varnames = [x for x in thekeys if
#                         (x[-len(str(dtset)):] == str(dtset) and not x[-len(str(dtset)) - 1:].isdigit())]
#             if len(varnames) > 0:
#                 varnames.sort()
#                 ret = ret + "#" + 60 * "-" + "\n#" + " DATASET " + str(dtset) + "\n#" + 60 * "-" + "\n\n"
#                 for i in varnames:
#                     if i == 'dmatpawu':
#                         if 2 in self['lpawu']:
#                             ret += self.write_key(i, ncolumns=5)
#                         elif 3 in self['lpawu']:
#                             ret += self.write_key(i, ncolumns=7)
#                     ret += self.write_key(i)
#                 ret += '\n'
#         return ret

#     def write_key(self, varname, ncolumns=None):
#         """
#         Receives an input variable and write their contents
#         properly according with their kind and length

#         Args:
#             varname:
#                 The name of the input variable
#         """

#         ret = ''
#         if varname not in self.variables:
#             print("[ERROR] input variable: '%s' is not defined" % varname)
#             return

#         # Assume that the variables are integer and test if such assumption
#         # is true
#         integer = True
#         real = False
#         string = False
#         compact = True

#         if isinstance(self.variables[varname], (int, float)):
#             varlist = [self.variables[varname]]
#         else:
#             varlist = self.variables[varname]

#         # Get the general kind of values for the input variable
#         for j in varlist:
#             try:
#                 if not float(j).is_integer():
#                     # This is the case of non integer values
#                     integer = False
#                     real = True
#                     string = False
#                     if len(str(float(j))) > 7:
#                         compact = False

#             except ValueError:
#                 # This is the case of '*1' that could not
#                 # be converted because we don't know the size
#                 # of the array
#                 integer = False
#                 real = False
#                 string = True

#         ret = ret + (varname.rjust(15)) + "  "

#         known_variables = {'xred': [3], 'acell': [3]}

#         if varname in known_variables:
#             for i in known_variables[varname]:
#                 if len(varlist) % i == 0:
#                     for j in range(int(len(varlist) / i)):
#                         if j == 0:
#                             ret += (i * '%17.10E ' + '\n') % tuple(varlist[j * i:j * i + i])
#                         else:
#                             ret += (17 * ' ' + i * '%17.10E ' + '\n') % tuple(varlist[j * i:j * i + i])
#         elif ncolumns is not None:
#             for i in ncolumns:
#                 for j in range(int(len(varlist) / i)):
#                     if j == 0:
#                         ret += (i * '%17.10E ' + '\n') % tuple(varlist[j * i:j * i + i])
#                     else:
#                         ret += (17 * ' ' + i * '%17.10E ' + '\n') % tuple(varlist[j * i:j * i + i])
#         else:
#             for j in range(len(varlist)):

#                 if real:
#                     if compact:
#                         ret += ("%g" % varlist[j]).rjust(8)
#                     else:
#                         ret += "%17.10e" % varlist[j]
#                 elif integer:
#                     ret += "%d" % varlist[j]
#                 elif string:
#                     ret += "%s" % varlist[j]

#                 # Conditions to jump to a new line
#                 if ((j + 1) % 3) == 0 and real and j < len(varlist) - 1:
#                     ret += "\n"
#                     ret += 17 * " "
#                 elif j < len(varlist) - 1:
#                     ret += " "
#             ret += "\n"
#         return ret


def merge(abi_into, abi_from, filename=None):
    """
    From the variables present in the file 'abi_into'
    will try to recover the value from 'abi_from'
    if the value has changed, the value will be updated.

    The new variables will be save in the given filename
    If filename==None the new values will overwrite
    abi_into

    Example:

       merge('abinit.in','abinit_xo_OUT.nc')

    It will update the abinit input with the values of the output

    :param abi_into: (str) Merge abinit variables using this destination
    :param abi_from:  (str) and this source
    :param filename: (str) Storing the final input variables on file

    """
    abinit_into = AbinitInput(abi_into)
    abinit_from = AbinitInput(abi_from)

    for i in abinit_into.variables.keys():
        if i in abinit_from.variables.keys():
            abinit_into.variables[i] = abinit_from.variables[i]

    if filename is None:
        filename = abi_into

    abinit_into.write(filename)


def xyz2input(filename):
    """
    Reads a .xyz and return an ABINIT input
    as a python dictionary
    """

    abiinput = AbinitInput()
    atomdict = atomic_symbol()
    rf = open(filename, 'r')

    natom = int(rf.readline())
    typat = []
    znucl = []
    xangst = []

    ntypat = 0
    rf.readline()
    data = rf.readlines()
    for i in range(natom):
        atom = data[i].split()
        atomnumber = atomdict[atom[0]]
        if atomnumber not in znucl:
            ntypat += 1
            znucl.append(atomnumber)
        typat.append(znucl.index(atomnumber) + 1)
        xangst += [float(atom[1]), float(atom[2]), float(atom[3])]

    abiinput.variables['natom'] = np.array([natom])
    abiinput.variables['znucl'] = np.array(znucl)
    abiinput.variables['ntypat'] = np.array([ntypat])
    abiinput.variables['typat'] = np.array(typat)
    abiinput.variables['xcart'] = angstrom_bohr * np.array(xangst)

    return abiinput



import os as _os
import shutil as _shutil
import numpy as _np


class OctopusStates:
    def __init__(self, basedir):
        rf = open(basedir + '/restart/gs/states')

        self.basedir = basedir
        self.nst = int(rf.readline().split()[1])
        self.dim = int(rf.readline().split()[1])
        self.nik = int(rf.readline().split()[1])

    def __str__(self):
        return "nst= %25d\ndim= %25d\nnik= %25d" % (self.nst, self.dim, self.nik)

    def __repr__(self):
        return "states('%s')" % self.basedir

    def write(self, filename):
        wf = open(filename, 'w')
        wf.write(self.__str__())
        wf.close()


class OctopusOccupancies:
    def __init__(self, basedir):
        rf = open(basedir + '/restart/gs/occs')

        self.basedir = basedir
        self.header = rf.readline() + rf.readline()
        rf.close()

        self.data = _np.loadtxt(basedir + '/restart/gs/occs', ndmin=1, skiprows=2, delimiter=' | ', comments='%',
                                dtype={'names': ['occupations', 'eigenvalue[a.u.]', 'k-points X', 'k-points Y',
                                                 'k-points Z', 'k-weights', 'filename', 'ik', 'ist', 'idim'],
                                       'formats': ['f8', 'f8', 'f8', 'f8', 'f8', 'f8', 'S10', 'i4', 'i4', 'i4']})

    def __repr__(self):
        return "occs('%s')" % self.basedir

    def __str__(self):
        text = self.header
        for line in self.data:
            text += "%21.14e | %21.14e | %21.14e | %21.14e | %21.14e | %21.14e | %10s | %3d | %3d | %3d\n" % tuple(line)
        text += "%"
        return text

    def write(self, filename):
        wf = open(filename, 'w')
        wf.write(self.__str__())
        wf.close()


class OctopusWavefunctions:
    def __init__(self, basedir):

        rf = open(basedir + '/restart/gs/wfns')

        self.basedir = basedir
        self.header = rf.readline() + rf.readline()
        rls = rf.readlines()
        lines = [x.split(' | ') for x in rls]
        nlines = len(lines) - 2
        rf.close()

        self.data = _np.zeros((nlines, 3))
        self.filenames = []
        for i in range(nlines):
            self.data[i, 0] = int(lines[i][0])
            self.data[i, 1] = int(lines[i][1])
            self.data[i, 2] = int(lines[i][2])
            self.filenames.append(lines[i][3].strip())
        self.iter = int(lines[-1][0].strip().split()[-1])

    def __repr__(self):
        return "wfns('%s')" % self.basedir

    def __str__(self):
        text = self.header
        for i in range(len(self.data)):
            text += "%8d | %8d | %8d | %12s\n" % tuple(list(self.data[i]) + [self.filenames[i]])
        text = text + '%\n' + ("Iter = %7d" % self.iter)
        return text

    def write(self, filename):
        wf = open(filename, 'w')
        wf.write(self.__str__())
        wf.close()


class OctopusStatesFile:
    def __init__(self, basedir):
        self.basedir = basedir
        self.obfs = sorted([x for x in _os.listdir(basedir + '/restart/gs')
                            if _os.path.isfile(basedir + '/restart/gs/' + x) and x[-4:] == '.obf' and x[0] == '0'])

    def __repr__(self):
        return "states_obf('%s')" % self.basedir


class OctopusRestart:
    def __init__(self, basedir):
        self.states = OctopusStates(basedir)
        self.occs = OctopusOccupancies(basedir)
        self.wfns = OctopusWavefunctions(basedir)
        self.states_obf = OctopusStatesFile(basedir)


def oct_merge(res1, res2, basedir, mksym1, mksym2):
    for i in [basedir, basedir + '/restart', basedir + '/restart/gs']:
        if not _os.path.isdir(i):
            _os.mkdir(i)

    res1.states.nst += res2.states.nst
    assert res1.states.dim == res2.states.dim
    assert res1.states.nik == res2.states.nik
    res1.states.write(basedir + '/restart/gs/states')

    lastf = int(res1.occs.data[-1][6])  # last filename in res1
    lasts = int(res1.occs.data[-1][8])  # last state number in res1
    lasts = res1.wfns.data[-1][1]  # last state number in res1
    lastf = int(res1.wfns.filenames[-1][1:-1])  # last filename in res1
    lastfilename = lastf
    index = 0
    for i in range(len(res2.occs.data)):
        res2.occs.data[i][6] = str(lastf + 1).zfill(10)
        res2.occs.data[i][8] = lasts + 1
        res2.wfns.data[i][1] = lasts + 1
        res2.wfns.filenames[i] = '"' + str(lastf + 1).zfill(10) + '"'
        lastf += 1
        if index == 0:
            index = 1
        else:
            index = 0
            lasts += 1
    res1.occs.data = _np.concatenate((res1.occs.data, res2.occs.data), axis=0)
    res1.occs.write(basedir + '/restart/gs/occs')
    res1.wfns.data = _np.concatenate((res1.wfns.data, res2.wfns.data), axis=0)
    res1.wfns.filenames += res2.wfns.filenames
    res1.wfns.write(basedir + '/restart/gs/wfns')

    # Copy or make symlinks of the States (*.obf) from first directory
    for i in res1.states_obf.obfs:
        print('FIRST: Copying or Linking ', i)
        if mksym1:
            _os.symlink(_os.getcwd() + '/' + res1.states_obf.basedir + '/restart/gs/' + i, basedir + '/restart/gs/' + i)
        else:
            _shutil.copy2(res1.states_obf.basedir + '/restart/gs/' + i, basedir + '/restart/gs')

    # Copy or make symlinks of the States (*.obf) from second directory
    for i in res2.states_obf.obfs:
        print('SECOND: Copying or Linking ', i)
        if mksym2:
            _os.symlink(_os.getcwd() + '/' + res2.states_obf.basedir + '/restart/gs/' + i,
                        basedir + '/restart/gs/' + str(lastfilename + 1).zfill(10) + '.obf')
        else:
            _shutil.copyfile(res2.states_obf.basedir + '/restart/gs/' + i,
                             basedir + '/restart/gs/' + str(lastfilename + 1).zfill(10) + '.obf')
        lastfilename += 1

    dirs = [x for x in _os.listdir(res1.states.basedir + '/restart/gs/') if
            x[0] != '0' and x not in ['states', 'occs', 'wfns']]

    # Copy the other files
    print(dirs)
    for i in dirs:
        _shutil.copy2(res1.states.basedir + '/restart/gs/' + i, basedir + '/restart/gs')


def merge_restart(dir1, dir2, dest, mksym1=True, mksym2=False):
    """
    Merge the restart directories (dir1 and dir2)
    into the destination (dest)
    """
    res1 = OctopusRestart(dir1)
    res2 = OctopusRestart(dir2)
    oct_merge(res1, res2, dest, mksym1, mksym2)



#!D:\bld\pychemia_1592382025033\_h_env\python.exe

import os
import sys
import json
import argparse
import numpy as np
import matplotlib
import multiprocessing
import logging

matplotlib.use('agg')
import matplotlib.pyplot as plt
from pychemia.code.abinit import AbinitInput, AbinitOutput
from pychemia.population.orbitaldftu import dmatpawu2params, params2dmatpawu, OrbitalDFTU
from pychemia.population.orbitaldftu import get_final_abinit_out
from pychemia.searcher import FireFly
from pychemia.db import get_database


def compare_params(path):
    if not os.path.isfile(path + os.sep + 'abinit.in'):
        print('ERROR: No abinit.in found at %s' % path)
        return
    # For making easier to see the values
    np.set_printoptions(linewidth=200, suppress=True)

    # Reading the INPUT
    abi = AbinitInput(path + os.sep + 'abinit.in')

    if 'lpawu' not in abi.variables:
        raise ValueError("Variable lpawu not found")
    ndim = 2 * max(abi['lpawu']) + 1

    idmatpawu = np.array(abi['dmatpawu']).reshape(-1, ndim, ndim)
    iparams = dmatpawu2params(idmatpawu, ndim)

    # Reading the OUTPUT
    abinitout = get_final_abinit_out(path)
    abo = AbinitOutput(abinitout)
    dmatpawu = abo.get_final_dmatpawu()
    odmatpawu = np.array(dmatpawu).reshape(-1, ndim, ndim)
    oparams = dmatpawu2params(odmatpawu, ndim)

    print('DMATPAWU')
    print('input')
    print(idmatpawu)
    print('output')
    print(odmatpawu)

    print('PARAMETRIC REPRESENTATION found at %s' % abinitout)
    for i in sorted(list(iparams.keys())):
        print(i)
        print('input')
        print(iparams[i])
        print('output')
        print(i)
        print(oparams[i])

    abo = AbinitOutput(abinitout)
    if not abo.is_finished:
        print('This output is not finished')
    try:
        nres2 = abo.get_energetics()['nres2'][-1]
        etot = abo.get_energetics()['etot'][-1]
        nscf = len(abo.get_energetics()['etot'])
        print("%30s ETOT: %15.6f NRES2: %15.6e NumSCF: %3d" % (path, etot, nres2, nscf))
    except:
        print("ERROR: Could not get energetics from %s" % abinitout)


def check_status(basepath):
    dirs = [x for x in os.listdir(basepath)
            if os.path.isdir(basepath + os.sep + x) and os.path.isfile(basepath + os.sep + x + os.sep + 'abinit.in')]
    print("%-40s %15s %15s %4s" % ("ABINIT output", "ETOT", 'nres2', 'nSCF'))
    for i in dirs:
        path = basepath + os.sep + i
        abinitout = get_final_abinit_out(path)
        if abinitout is None:
            continue

        abo = AbinitOutput(abinitout)
        if not abo.is_finished:
            continue
        try:
            nres2 = abo.get_energetics()['nres2'][-1]
            etot = abo.get_energetics()['etot'][-1]
            nscf = len(abo.get_energetics()['etot'])
            print("%-40s %15.6f %15.6e %4d" % (abinitout, etot, nres2, nscf))
        except:
            print("ERROR: Could not get final energetics from %s" % abinitout)


def plot_polar(popu, basepath):
    print('Plotting Euler Angles...')

    dirs = [x for x in os.listdir(basepath)
            if os.path.isdir(basepath + os.sep + x) and os.path.isfile(basepath + os.sep + x + os.sep + 'abinit.in')]
    fig = plt.figure(figsize=(21, 1.2 * len(dirs)))
    plt.subplots_adjust(left=0.0, bottom=0.0, right=0.99, top=0.99, wspace=None, hspace=None)
    etots = []

    for idir in dirs:
        pm, etot = popu.get_final_properties(basepath + os.sep + idir)
        etots.append(etot)

    etots = np.array(etots)
    sort_dirs = np.array(dirs)[etots.argsort()]

    index = 0.0
    for idir in sort_dirs:

        pm, etot = popu.get_final_properties(basepath + os.sep + idir)
        ea = np.array(pm['final_dmat']['euler_angles'])
        # print(idir,etot)
        # print(ea.shape)
        nangles = ea.shape[1]

        for j in range(nangles):
            theta = ea[:, j]
            dim = len(theta)
            radii = np.ones(dim)
            colors = np.arange(dim)
            width = 0.1 * np.ones(dim)

            ax = fig.add_axes([float(j) / nangles, index / (len(dirs) + 1), 1.0 / nangles, 1.0 / nangles],
                              projection='polar')
            ax.yaxis.set_tick_params(labelsize=0)
            ax.xaxis.set_tick_params(labelsize=0)
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)
            ax.spines['polar'].set_visible(True)
            bars = ax.bar(theta, radii, width=width, bottom=0.0)

            # Use custom colors and opacity
            for r, bar in zip(colors, bars):
                bar.set_facecolor(plt.cm.viridis(float(r) / nangles))
                bar.set_alpha(0.9)

        index += 1.0

    plt.savefig('OrbitalPolar.pdf')
    plt.show()


def plot_status(basepath):
    dirs = [x for x in os.listdir(basepath)
            if os.path.isdir(basepath + os.sep + x) and os.path.isfile(basepath + os.sep + x + os.sep + 'abinit.in')]

    abi = AbinitInput(basepath + os.sep + 'abinit.in')
    nstep = abi['nstep']

    print('Plotting Status...')
    xx = {}
    yy = {}

    # Get the data:
    max_nruns = 0
    for path in dirs:
        xx[path] = np.array([])
        yy[path] = {}
        yy[path]['nres2'] = np.array([])
        yy[path]['etot'] = np.array([])
        yy[path]['delta'] = np.array([])

        for i in range(100):
            if os.path.isfile(basepath + os.sep + path + os.sep + 'abinit_%02d.txt' % i):
                abo = AbinitOutput(basepath + os.sep + path + os.sep + 'abinit_%02d.txt' % i)
                if not abo.is_finished:
                    print('This output is not finished')
                    continue
                if max_nruns < i:
                    max_nruns = i
                try:
                    energ = abo.get_energetics()
                except:
                    raise RuntimeError(
                        "Failed procesing: %s" % (basepath + os.sep + path + os.sep + 'abinit_%02d.txt' % i))

                nres2 = energ['nres2'][-1]
                etot = energ['etot'][-1]
                nscf = len(energ['etot'])
                x = np.arange(nstep * i, nstep * i + len(energ['nres2']))
                yetot = np.array(energ['etot'])
                ynres2 = np.array(energ['nres2'])
                ydelta = np.array(np.abs(energ['deltaEh']))
                xx[path] = np.concatenate((xx[path], x))
                yy[path]['etot'] = np.concatenate((yy[path]['etot'], yetot))
                yy[path]['nres2'] = np.concatenate((yy[path]['nres2'], ynres2))
                yy[path]['delta'] = np.concatenate((yy[path]['delta'], ydelta))
                print("%s ETOT:%15.6f NRES2=%15.6e Num SCF=%3d" % (path, etot, nres2, nscf))

    # RESIDUAL
    plt.figure(figsize=(8, 11))
    miny = 1E-1
    maxy = 1E-16
    plt.subplots_adjust(left=0.1, bottom=0.05, right=0.99, top=0.99,
                        wspace=None, hspace=None)
    for path in dirs:
        if len(xx) > 0:
            plt.semilogy(xx[path], yy[path]['nres2'], label=path[-4:], lw=0.1)
            if miny > min(yy[path]['nres2']):
                miny = min(yy[path]['nres2'])
            if maxy < max(yy[path]['nres2']):
                maxy = max(yy[path]['nres2'])

    plt.ylim(miny, maxy)

    for i in nstep * np.arange(max_nruns + 1):
        plt.semilogy([i, i], [miny, maxy], '0.5', lw=0.1)

    plt.xlim(0, max([max(xx[path]) for path in dirs]))
    plt.legend()
    plt.xlabel("SCF iteration")
    plt.ylabel("Density Residual$^2$")
    plt.savefig('Orbital_Residual.pdf')

    # ETOT
    miny = 1E6
    maxy = -1E6
    avg = 0
    minlen = 100
    plt.figure(figsize=(8, 11))
    for path in dirs:
        if len(xx) > 0:
            plt.plot(xx[path], yy[path]['etot'], label=path[-4:])
            if len(yy[path]['etot']) < minlen:
                minlen = len(yy[path]['etot'])
                avg = np.average(yy[path]['etot'][-int(minlen / 2):])
            if miny > min(yy[path]['etot'][-int(minlen / 2):]):
                miny = min(yy[path]['etot'][-int(minlen / 2):])
            if maxy < max(yy[path]['etot'][-int(minlen / 2):]):
                maxy = max(yy[path]['etot'][-int(minlen / 2):])

    plt.subplots_adjust(left=0.15, bottom=0.05, right=0.95, top=0.95,
                        wspace=None, hspace=None)

    newminy = miny - 0.1 * (maxy - miny)
    newmaxy = maxy + 0.1 * (maxy - miny)
    miny = newminy
    maxy = newmaxy

    plt.ylim(miny, maxy)
    for i in nstep * np.arange(max_nruns + 1):
        plt.plot([i, i], [miny, maxy], '0.5', lw=0.1)

    plt.xlim(0, max([max(xx[path]) for path in dirs]))
    plt.legend()
    plt.xlabel("SCF iteration")
    plt.ylabel("Energy")
    plt.savefig('Orbital_ETotal.pdf')

    # deltaEh
    plt.figure(figsize=(8, 11))
    miny = 1E-1
    for path in dirs:
        if len(xx) > 0:
            plt.semilogy(xx[path], yy[path]['delta'], label=path[-4:], lw=0.1)
            if miny > min(yy[path]['delta']):
                miny = min(yy[path]['delta'])

    plt.subplots_adjust(left=0.1, bottom=0.05, right=0.99, top=0.99,
                        wspace=None, hspace=None)

    for i in nstep * np.arange(max_nruns + 1):
        plt.semilogy([i, i], [miny, 1E3], '0.5', lw=0.1)

    plt.ylim(miny, 1E3)
    plt.xlim(0, max([max(xx[path]) for path in dirs]))
    plt.legend()
    plt.xlabel("SCF iteration")
    plt.ylabel("Delta Energy")
    plt.savefig('Orbital_deltaEh.pdf')


def create_population(num_candidates):
    popu.random_population(num_candidates)


def safe_read_json(filename):
    """
    Safely read a given filename, extract and returns the associated dictionary
    from the JSON file
    """
    if not os.path.exists(filename):
        raise ValueError("ERROR: Could not read file: %s" % filename)
    rf = open(filename)
    try:
        data = json.load(rf)
    except ValueError:
        raise ValueError("ERROR: File is not in proper JSON format: %s" % filename)
    rf.close()
    return data


def prepare_folders(scrpath):
    for i in popu.members:
        popu.prepare_folder(i, workdir=scrpath, source_dir=scrpath)


def set_populations(parser):

    return parser.add_argument('-p', type=str, nargs='+', required=True, metavar='JSON_file',
                               help='Population settings (JSON file), includes settings to connect to server and the '
                                    'population')


def evaluate(db_settings, queue_settings, abipath):
    pcdb = get_database(db_settings)
    print("[%s] Path for abinit.in: %s" % (pcdb.name, abipath))
    popu = OrbitalDFTU(pcdb, abipath + os.sep + 'abinit.in')
    popu.evaluator(queue_settings, abipath)


def search(db_settings, search_settings, abipath):
    pcdb = get_database(db_settings)
    print("[%s] Path for abinit.in: %s" % (pcdb.name, abipath))
    popu = OrbitalDFTU(pcdb, abipath + os.sep + 'abinit.in')

    if 'generation_size' in search_settings:
        generation_size = search_settings.pop('generation_size')
    else:
        generation_size = 32

    if 'stabilization_limit' in search_settings:
        stabilization_limit = search_settings.pop('stabilization_limit')
    else:
        stabilization_limit = 10

    fire = FireFly(popu, params=search_settings, generation_size=generation_size,
                   stabilization_limit=stabilization_limit)
    fire.run()


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Orbital DFTU Evaluator and Analysis Tool')

    subparsers = parser.add_subparsers(help='commands', dest='subparser_name')

    # The create command
    create_parser = subparsers.add_parser('create', help='Create the database')

    set_populations(create_parser)

    # The populate command
    populate_parser = subparsers.add_parser('populate', help='Add candidates to the population (used for testing)')

    set_populations(populate_parser)

    populate_parser.add_argument('-clean', action='store_true',
                                 help='If true, clean the entire database before populate',
                                 required=False, default=False)

    populate_parser.add_argument('-size', type=int,
                                 help='Number of candidates to populate (default: 16)',
                                 required=False, default=16)
    populate_parser.add_argument('-basepath', type=str,
                                 help='Path where calculations are performed',
                                 required=False, default='.')

    # A run command
    run_parser = subparsers.add_parser('run', help='Run Evaluator')

    set_populations(run_parser)

    run_parser.add_argument('-queue_settings', type=str,
                            help='Filename with PBS settings for launching jobs',
                            required=False, default='queue.json')
    run_parser.add_argument('-basepath', type=str,
                            help='Path where calculations are performed (default: current folder)',
                            required=False, default='.')

    # A searcher command
    searcher_parser = subparsers.add_parser('search', help='Run PyChemia Global Searcher')

    set_populations(searcher_parser)

    searcher_parser.add_argument('-search_settings', type=str,
                                 help='Filename with PBS settings for launching jobs',
                                 required=False, default='searcher.json')
    searcher_parser.add_argument('-basepath', type=str,
                                 help='Path where calculations are performed (default: current folder)',
                                 required=False, default='.')

    # The plot command
    plot_parser = subparsers.add_parser('plot', help='Generate several plots')

    set_populations(plot_parser)

    plot_parser.add_argument('-basepath', type=str,
                             help='Path where calculations are performed',
                             required=False, default='.')

    args = parser.parse_args()
    print(args)

    # check all settings in args.p
    dbs = []
    for dbi_file in args.p:
        dbi = safe_read_json(dbi_file)
        print("DB settings: %s" % dbi)
        assert('name' in dbi)
        assert('u' in dbi)
        assert('j' in dbi)
        dbs.append(dbi)

    if args.subparser_name == 'create':
        
        pcdbs = []
        for dbi in dbs:
            pcdb = get_database(dbi)
            pcdbs.append(pcdb)
            print(pcdb)
            print(pcdb.entries.count())

    if args.subparser_name == 'run':
        queue_settings = safe_read_json(args.queue_settings)
        print("PBS settings: %s" % queue_settings)

    # General actions for 'populate', 'run', 'search' and 'plot'
    if args.subparser_name in ['run', 'plot', 'populate', 'search']:

        if not os.path.isdir(args.basepath) or not os.path.isfile(args.basepath + os.sep + 'abinit.in'):
            print('ERROR: Wrong basepath %s, directory must exist and contain a abinit.in file' % args.basepath)
            parser.print_help()
            sys.exit(1)

        popu = {}
        for dbi in dbs:
            name = dbi['name']
            pcdb = get_database(dbi)
            if not os.path.isdir(args.basepath + os.sep + name):
                os.mkdir(args.basepath + os.sep + name)
            abi = AbinitInput(args.basepath + os.sep + 'abinit.in')
            
            abi['upawu'] = ""
            for i in dbi['u']:
                abi['upawu'] += str(i) + " "
            abi['upawu'] += 'eV'
                
            abi['jpawu'] = ""
            for i in dbi['j']:
                abi['jpawu'] += str(i) + " "
            abi['jpawu'] += 'eV'

            abipath = args.basepath + os.sep + name + os.sep + 'abinit.in'
            abi.write(abipath)

            abifiles = args.basepath + os.sep + name + os.sep + 'abinit.files'
            if os.path.lexists(abifiles):
                os.remove(abifiles)
            os.symlink(os.path.abspath(args.basepath + os.sep + 'abinit.files'), abifiles)
            for i in [x for x in os.listdir(args.basepath) if x[-3:] == 'xml']:
                psp = args.basepath + os.sep + name + os.sep + i
                if os.path.lexists(psp):
                    os.remove(psp)
                os.symlink(os.path.abspath(args.basepath + os.sep + i), psp)

            popu[name] = OrbitalDFTU(pcdb, abipath)

    # Specific actions for 'populate', 'run', 'search' and 'plot'
    if args.subparser_name == 'populate':

        for dbi in dbs:
            name = dbi['name']
            if args.clean:
                popu[name].clean()
            popu[name].random_population(args.size)
            print("[%s] Total number of candidates: %d" % (name, len(popu[name])))

    elif args.subparser_name == 'run':
        sprocs = {}
        for dbi in dbs:
            name = dbi['name']
            basepath = args.basepath + os.sep + name

            if os.path.exists(basepath + os.sep + 'ERROR'):
                print('ERROR: Something was wrong with %s' % abipath)
            else:
                sprocs[name] = multiprocessing.Process(target=evaluate, args=(dbi, queue_settings, basepath))
                sprocs[name].start()

        for dbi in dbs:            
            name = dbi['name']
            sprocs[name].join()

    elif args.subparser_name == 'search':
        logging.basicConfig(level=logging.DEBUG)
        search_settings = safe_read_json(args.search_settings)
        print("Search settings from file: %s \n%s" % (args.search_settings, search_settings))

        sprocs = {}
        for dbi in dbs:
            name = dbi['name']
            basepath = args.basepath + os.sep + name
            sprocs[name] = multiprocessing.Process(target=search, args=(dbi, search_settings, basepath))
            sprocs[name].start()

        print(list(sprocs.keys()))

        for dbi in dbs:
            name = dbi['name']
            if name not in sprocs:
                print('ERRROR: %s' % str(sprocs))
            sprocs[name].join()

    elif args.subparser_name == 'plot':
        check_status(args.basepath)
        plot_status(args.basepath)
        plot_polar(popu, args.basepath)



# import pyximport
# pyximport.install()

import numpy as np
import scipy.optimize
import scipy.spatial
from pychemia import pcm_log
from pychemia.utils.mathematics import length_vectors
from pychemia.utils.periodic import mass

from .lj_utils import lj_energy, lj_forces, lj_gradient





class LennardJones:
    def __init__(self, structure, ljparams=None, cp=0.0):
        self.initial_structure = structure
        self.structure = structure.copy()

        if ljparams is None:
            ljparams = {}
            for i in self.structure.species:
                for j in self.structure.species:
                    ljparams[i + "-" + j] = {"sigma": 1.0, "epsilon": 1.0}
        self.ljparams = ljparams
        self.cp = cp
        self.sigmas = np.zeros((structure.natom, structure.natom))
        self.epsilons = np.zeros((structure.natom, structure.natom))
        for i in range(structure.natom):
            for j in range(structure.natom):
                self.sigmas[i, j], self.epsilons[i, j] = self.get_sigma_epsilon(i, j)

    def get_forces(self):
        return lj_forces(
            self.structure.positions, self.sigmas, self.epsilons, self.cp
        ).reshape((-1, 3))

    def get_magnitude_forces(self):
        return length_vectors(self.get_forces())

    def steepest_descent(self, dt=1, tolerance=1e-3):
        forces = self.get_forces()
        while np.max(self.get_magnitude_forces()) > tolerance:
            for i in range(self.structure.natom):
                imass = mass(self.structure.symbols[i])
                self.structure.positions[i] += 0.5 * forces[i] / imass * dt ** 2

            forces = self.get_forces()
            maxforce = np.max(self.get_magnitude_forces())
            dt = max(100.0 / maxforce, 1.0)
            pcm_log.debug(
                "Steepest Descent maxforce= %9.3E   density=%7.4f dt=%7.3f"
                % (maxforce, self.structure.density, dt)
            )

    def get_sigma_epsilon(self, i, j):
        specie1 = self.structure.symbols[i]
        specie2 = self.structure.symbols[j]
        sort_species = [specie1, specie2]
        sort_species.sort()

        sigma = self.ljparams[sort_species[0] + "-" + sort_species[1]]["sigma"]
        epsilon = self.ljparams[sort_species[0] + "-" + sort_species[1]]["epsilon"]

        return float(sigma), float(epsilon)

    def get_energy(self):
        return lj_energy(self.structure.positions, self.sigmas, self.epsilons, self.cp)

    def local_minimization(
        self, method="BFGS", gtol=1e-4, soft_max_ncalls=5, hard_max_ncalls=20
    ):

        if method in ["Nelder-Mead", "Powell"]:
            jac = None
            options = {"maxiter": 100, "disp": False}
        else:
            jac = lj_gradient
            options = {"gtol": 0.1 * gtol, "disp": False}

        x0 = self.structure.positions.flatten()

        hard_call = 1
        soft_call = 1
        res = 0.0

        while True:

            res = scipy.optimize.minimize(
                lj_energy,
                x0,
                args=(self.sigmas, self.epsilons, self.cp),
                method=method,
                jac=jac,
                options=options,
            )

            if method in ["BFGS", "CG"]:
                maxforce = np.max(
                    np.apply_along_axis(
                        np.linalg.norm, 1, np.array(res.jac).reshape((-1, 3))
                    )
                )
                pcm_log.debug(
                    "[%2d/%2d] MaxForce/gtol= %9.3E/%9.3E  Value= %12.5f"
                    % (soft_call, soft_max_ncalls, maxforce, gtol, res.fun)
                )

                x0 = res.x
                soft_call += 1
                hard_call += 1
                if maxforce < gtol:
                    break
                if soft_call > soft_max_ncalls:
                    pcm_log.debug(
                        "Anomalous condition, maxforce= %9.3E gtol= %9.3E"
                        % (maxforce, gtol)
                    )
                    pos = np.array(x0).reshape((-1, 3))
                    mindis = np.min(
                        np.array(
                            np.array(scipy.spatial.distance_matrix(pos, pos))
                            + 100 * np.eye(len(pos))
                        ).flatten()
                    )
                    x0 = np.array(1.0 / mindis * pos).flatten()
                    soft_call = 1
                if hard_call > hard_max_ncalls:
                    pcm_log.debug(
                        "Could not reach target forces, maxforce= %9.3E gtol= %9.3E"
                        % (maxforce, gtol)
                    )
                    pcm_log.debug("Hard limit reached, aborting local minimization")
                    break
            else:
                break
        return res

    def move_to_minima(self):
        relax = self.local_minimization()
        self.structure.positions = relax.x.reshape((-1, 3))
        self.structure.relocate_to_cm()


def lj_compact_evaluate(structure, gtol, minimal_density):
    print(structure)
    k = 1
    while structure.density < minimal_density:
        iniden = structure.density
        lj = LennardJones(structure, cp=k)
        relax = lj.local_minimization(gtol=gtol)
        structure.set_positions(relax.x.reshape((-1, 3)))
        finden = structure.density
        pcm_log.debug(
            "Compacting Cluster (target_density=%7.3f): Density I= %7.3f   F= %7.3f"
            % (minimal_density, iniden, finden)
        )
        k += 1
        if k > 10:
            pcm_log.debug("I tried too much...")
            break

    lj = LennardJones(structure)
    relax = lj.local_minimization(gtol=gtol)
    # Return relaxed positions, forces and energy
    return relax.x.reshape((-1, 3)), relax.jac.reshape((-1, 3)), relax.fun



"""
One core capability of PyChemia is global optimization of atomic structures such as clusters and 
crystals. Doing global optimization requires compute forces and energies of hundreds or even 
thousands of structures and all these data must be store and processed efficiently.

PyChemia relies of MongoDB to store structures and the properties computed by atomistic codes.
This module creates and manipulates Mongo databases.
There are two kinds of databases defined on PyChemia: __PyChemiaDB__ is a kind of database to store 
structure and properties. __PyChemiaQueue__ is a repository of calculations.

In the case of Global searcher PyChemiaDB contains several collections, such as:

* pychemia_entries: The main collection, contains all the structures, properties and status info

* fingeprints: Collection with using the same IDs as pychemia_entries storing the fingerprints of the corresponding
  structures

* generations: Collection with the same IDs as pychemia_entries to store on which generation each structure belongs.

* generation_changes: Collection storing the changes introduced from one generation to the next one.

* population_info: Stores the basic parameters entered to create the population

* searcher_info: Store parameters about the searcher use to populate the database

* lineage: Store the heritage of structures produced by the global searcher

To do a more efficient executiion of calculations using several computer clusters and dedicated 
machines. PyChemia provides a central database of calculations that can be used to feed clusters 
with execution jobs. The PyChemiaQueue is used to fill the role of a meta-queue for structures 
and jobs that need to be computed.
"""

from pychemia import HAS_PYMONGO, HAS_GRIDFS

if HAS_PYMONGO:
    from .db import PyChemiaDB, get_database, object_id, create_database, has_connection

    if HAS_GRIDFS:
        from .queue import PyChemiaQueue

# __all__ = filter(lambda s: not s.startswith('_'), dir())



from math import sqrt as _sqrt
import pychemia.utils.constants as _pc
from pychemia.utils.computing import string2number

"""
A syntactic parser for ABINIT input files ".in"
"""

__author__ = "Guillermo Avendano-Franco"
__copyright__ = "Copyright 2016"
__version__ = "1.1"
__maintainer__ = "Guillermo Avendano-Franco"
__email__ = "guillermo.avendano@uclouvain.be"
__status__ = "Development"
__date__ = "May 13, 2016"


def __parse_word(word):
    """
    Read an string and determines if it is a word or a number
    or a repetitive list os values such as 3*4.5=[4.5 , 4.5, 4.5]

    Args:
        word: An string that should be parsed

    Returns:
       result:
          Value extracted (word, number or list)

       kind:
          'word' such as 'ntime', 'eV', '*1', 'angstrom', etc
          'int' such as 1, 2, 3
          'float' such as 4.5, 6.7, etc
          'list' such as [4.5 , 4.5, 4.5]

    """
    result = None
    kind = None
    if word[0].isalpha() and word[:4] != 'sqrt' and word[:5] != '-sqrt':
        result = word
        kind = 'word'

    elif word[:4] == 'sqrt':
        result = _sqrt(float(word[5:-1]))
        kind = 'float'

    elif word[:5] == '-sqrt':
        result = -_sqrt(float(word[6:-1]))
        kind = 'float'

    elif word[0] == '*':
        result = word
        kind = 'word'

    elif word.isdigit():
        result = int(word)
        kind = "int"

    elif '*' in word:
        splt = word.split('*')

        if splt[0].isdigit():
            mult = int(splt[0])
            number, kind = string2number(splt[1])
            if number is not None:
                result = mult * [number]
                kind = 'list'
            else:
                result = None
                kind = None
    else:
        result, kind = string2number(word)

    return result, kind


def __clean_line(line):
    """
    Removes comments from a line, the comments
    could be with '#' or '!'
    Both kind of comments are used in ABINIT
    """
    line = line.strip()
    if len(line) > 0:
        if line[0] == '#' or line[0] == '!':
            result = []
        elif '#' in line:
            splt = line.split('#')
            line = splt[0]
            if '!' in line:
                splt = line.split('!')
                line = splt[0]
            result = line.split()
        elif '!' in line:
            splt = line.split('!')
            line = splt[0]
            if '#' in line:
                splt = line.split('#')
                line = splt[0]
            result = line.split()
        else:
            result = line.split()
    else:
        result = []
    return result


def __clean_input(filename):
    """
    Take an input file and return an
    string with all the comments removed

    """
    rf = open(filename, 'r')
    result = []
    for line in rf.readlines():
        ans = __clean_line(line)
        result = result + ans
    rf.close()
    return result


def parser(filename):
    """
    Take a filename and returns
    a dictionary of ABINIT input variables

    The values on each key are lists of numbers
    The values are converted to atomic units always
    that it found other variables (eV, angstrom, etc)
    Repetitions are expanded (3*1.5 => [1.5 1.5 1.5])

    Args:
       filename:
          Path for an ABINIT input file

    Returns:
       inputvars:
          Dictionary, the keys are names
          of variables and the values are
          list of numbers.

    """
    cleanlist = __clean_input(filename)

    inputvars = {}
    keyword = cleanlist[0]
    result, kind = __parse_word(keyword)
    if kind != 'word':
        raise ValueError('ERROR(parser): The first element', keyword, 'should be a keyword\n' + filename)
    varlist = []
    for word in cleanlist[1:]:
        value, kind = __parse_word(word)
        if kind == 'word':
            assert (isinstance(value, str))
            value = str(value)
            # It could be:
            # new keyword such as ntime
            # scaling factor as eV, ha
            # indefined multiplicator such as '*1'
            if value.lower()[:6] == "angstr":
                if len(varlist) == 0:
                    print('ERROR: scaling factor', value, 'before a list of values')
                else:
                    varlist = [_pc.angstrom_bohr * x for x in varlist]
            elif value.lower() == "ev":
                if len(varlist) == 0:
                    print('ERROR: scaling factor', value, 'before a list of values')
                else:
                    varlist = [_pc.eV_Ha * x for x in varlist]
            elif value.lower() == "ry":
                if len(varlist) == 0:
                    print('ERROR: scaling factor', value, 'before a list of values')
                else:
                    varlist = [0.5 * x for x in varlist]
            elif value.lower() == "t" or value.lower() == "te":
                if len(varlist) == 0:
                    print('ERROR: scaling factor', value, 'before a list of values')
                else:
                    varlist = [_pc.BField_Tesla * x for x in varlist]
            elif value.lower() == "k":
                if len(varlist) == 0:
                    print('ERROR: scaling factor', value, 'before a list of values')
                else:
                    varlist = [_pc.kb_HaK * x for x in varlist]
            elif value.lower() == "ha" or value.lower() == "hartree":
                pass
            elif value.lower() == "bohr" or value.lower() == "au":
                pass
            elif value[0] == '*':
                varlist.append(value)
            elif keyword == 'xyzfile':
                print('FOUND case XYZFILE')
                varlist.append(value)
            else:
                # New keyword
                if len(varlist) == 0:
                    print('ERROR: new keyword "%s" with no elements for previos keyword "%s"' % (value, keyword))
                else:
                    # Store the varlist in the dictionary
                    # for the previuos keyword
                    if len(varlist) == 1:
                        inputvars[keyword] = varlist[0]
                    else:
                        inputvars[keyword] = varlist
                    # New keyword
                    keyword = value
                    varlist = []
        elif kind == 'int':
            varlist.append(value)
        elif kind == 'float':
            varlist.append(value)
        elif kind == 'list':
            varlist = varlist + value
        else:
            return None

    # Final keyword
    if len(varlist) == 0:
        print('ERROR: new keyword "%s" with no elements for previos keyword "%s" ' % (value, keyword))
    else:
        # Store the varlist in the dictionary
        # for the previuos keyword
        if len(varlist) == 1:
            inputvars[keyword] = varlist[0]
        else:
            inputvars[keyword] = varlist

    return inputvars



from .searcher import Searcher
from pychemia import pcm_log


class GreyWolf(Searcher):
    def __init__(self, population, params=None, generation_size=32, stabilization_limit=10):

        Searcher.__init__(self, population, generation_size, stabilization_limit)
        self.a = None
        self.c = None
        self.set_params(params)

    def set_params(self, params):
        self.a = 1
        self.c = 1
        if params is None:
            params = {}
        if 'a' in params:
            self.a = params['a']
        if 'c' in params:
            self.c = params['c']

    def get_params(self):
        return {'a': self.a, 'c': self.c}

    def run_one(self):
        # Get a static selection of the values in the generation that are relaxed
        selection = self.population.ids_sorted(self.population.actives_evaluated)
        pcm_log.info('Size of selection : %d' % len(selection))



import math

from .searcher import Searcher
from pychemia import pcm_log


class ParticleSwarm(Searcher):
    def __init__(self, population, params=None, generation_size=32, stabilization_limit=10):
        """
        Implementation fo the Firefly algorithm for global minimization
        This searcher uses a metric to compute the attractiveness and the vector displacement
        to move one firefly in the direction of another one

        :param population:
        :param params: (dict) Parameters to setup the Searcher
        :param generation_size: (int)
        :param stabilization_limit: (int)
        :return:
        """
        # Mandatory objects
        self.population = population
        # Parameters
        self.gamma = None
        self.elites = None
        self.set_params(params)
        # Constrains
        self.generation_size = generation_size
        self.stabilization_limit = stabilization_limit
        # Initializing objects
        Searcher.__init__(self, self.population, generation_size, stabilization_limit)

    def set_params(self, params):
        if params is None:
            self.gamma = 0.1
            self.elites = 3
        else:
            assert ('gamma' in params)
            assert (params['gamma'] >= 0.0)
            self.gamma = params['gamma']
            if 'elites' in params:
                self.elites = params['elites']

    def get_params(self):
        return {'gamma': self.gamma, 'elites': self.elites}

    def run_one(self):
        # Get a static selection of the values in the generation that are relaxed
        selection = self.population.ids_sorted(self.actives_in_generation)

        # Minus sign because we are searching for minima
        intensity = self.population.get_values(selection)
        for entry_id in intensity:
            intensity[entry_id] *= -1

        moves = {}
        new_selection = {}
        for entry_id in selection:
            new_selection[entry_id] = None

        # Move all the fireflies (Except the most brightness)
        # as the selection is sorted it means that the first one will no move
        pcm_log.debug('No Moving %d %s. Intensity: %7.3f' % (0, str(selection[0]), intensity[selection[0]]))

        # The best
        elites = selection[:self.elites]

        for i in range(self.elites, len(selection)):
            entry_id = selection[i]
            pcm_log.debug('Moving %d %s. Intensity: %7.3f' % (i, str(entry_id), intensity[entry_id]))

            distances = [self.population.distance(entry_id, entry_jd) for entry_jd in elites]
            target = elites[distances.index(min(distances))]
            distance = min(distances)
            atractiveness = math.exp(-self.gamma * distance) * intensity[target]

            pcm_log.debug('[%s] Distance: %7.3f. Intensity: %7.3f. Atractiveness: %7.3f' % (str(target),
                                                                                            distance,
                                                                                            intensity[target],
                                                                                            atractiveness))

            if intensity[entry_id] < atractiveness:
                new_selection[entry_id] = self.population.move(entry_id, target, in_place=False)

        for entry_id in selection:
            pcm_log.debug('Deciding fate for firefly: %s' % str(entry_id))
            if new_selection[entry_id] is not None:
                pcm_log.debug('Moved to a new location %s ' % str(entry_id))
                self.replace_by_other(entry_id, new_selection[entry_id], reason=None)
            else:
                pcm_log.debug('Promoted to new generation ')
                self.pass_to_new_generation(entry_id, reason='No other firefly is more attractive')



"""
Classes to manipulate execution on Queue systems, 'Torque'
"""

from ._runner import Runner
from ._pbs import PBSRunner, report_cover, get_jobs

# __all__ = filter(lambda s: not s.startswith('_'), dir())



# Real physical constants
# Revised fundamental constants from http://physics.nist.gov/cuu/Constants/index.html
# (from 2006 least squares adjustment)

#: 1 Bohr, in Angstrom
bohr_angstrom = 0.52917720859
angstrom_bohr = 1.0 / bohr_angstrom

#: 1 Hartree, in cm^-1
Ha_cmm1 = 219474.6313705

#: 1 Hartree, in eV
Ha_eV = 27.21138386
eV_Ha = 1 / Ha_eV

#: 1 Hartree, in meV
Ha_meV = Ha_eV * 1000

#: 1 Hartree, in Kelvin
Ha_K = 315774.65

# 1 Hartree, in THz
Ha_THz = 6579.683920722

#: 1 Hartree, in J
Ha_J = 4.35974394E-18

#:  Minus the electron charge, in Coulomb
e_Cb = 1.602176487E-19

#: Boltzmann constant in Ha/K
kb_HaK = 8.617343E-5 / Ha_eV

#: 1 atomic mass unit, in electronic mass
amu_emass = 1.660538782E-27 / 9.10938215E-31

#: 1 Ha/Bohr^3, in GPa
HaBohr3_GPa = Ha_eV / bohr_angstrom ** 3 * e_Cb * 1.0E+21

#: Avogadro number
Avogadro = 6.02214179E+23

#: Inverse of fine structure constant
InvFineStruct = 137.035999679

#: Speed of light in atomic units
Sp_Lt = 2.99792458E+8 / 2.1876912633E+6

#: Atomic unit of time, in seconds
Time_Sec = 2.418884326505E-17

#: Atomic unit of induction field (in Tesla) * mu_B (in atomic units).
BField_Tesla = 4.254383E-6 * 0.5
